import enum
import sys
from typing import List, Optional, Union

import torch


class Unit(enum.Enum):
    """
    Enum type for the different units understood by the program.
    """

    ARBITRARY = enum.auto()
    CM = enum.auto()
    MM = enum.auto()


class ModelOutput:
    """
    Base class packaging the raw Tensor output from a deep neural network
    with information about its unit and potentially the way it parameterizes
    a (batch of) response distributions, if probabilistic.
    """

    # Should all be implemented by subclasses
    computes_calibration = False
    ndim: int
    nnode: int

    def __init__(
        self,
        raw_output,
        arena_dims: torch.Tensor,
        arena_dim_units: Union[str, Unit],
    ):
        """Initialize a ModelOutput object."""
        self.raw_output = raw_output
        if isinstance(raw_output, torch.Tensor):
            self.batch_size = raw_output.shape[0]
            self.device = raw_output.device
        else:
            self.device = raw_output[0].raw_output.device

        arena_dims = arena_dims.to(self.device)
        self.ndim = len(arena_dims)

        if isinstance(arena_dim_units, str):
            try:
                arena_dim_units = Unit[arena_dim_units]
            except KeyError:
                raise ValueError(
                    "Arena dimensions must be provided in either centimeters or millimeters! "
                    f"Instead encountered unit: {arena_dim_units}."
                )

        self.arena_dims = {
            Unit.MM: self._convert(arena_dims, arena_dim_units, Unit.MM),
            Unit.CM: self._convert(arena_dims, arena_dim_units, Unit.CM),
        }

    def _convert(
        self, x: torch.Tensor, in_units: Unit, out_units: Unit
    ) -> torch.Tensor:
        """
        Convert an input array-like representing a (collection of) point(s)
        from one unit to another.
        """
        if in_units == out_units:
            return x
        elif in_units == Unit.MM and out_units == Unit.CM:
            return x / 10
        elif in_units == Unit.CM and out_units == Unit.MM:
            return x * 10
        elif in_units == Unit.ARBITRARY:
            dims = self.arena_dims[out_units].to(x.device)
            # This ensures the relative scale of each dimension is preserved
            scale_factor = dims.max() / 2
            return x * scale_factor
        elif out_units == Unit.ARBITRARY:
            dims = self.arena_dims[in_units].to(x.device)
            # This ensures the relative scale of each dimension is preserved
            scale_factor = dims.max() / 2
            return x / scale_factor
        else:
            raise ValueError(
                "Expected both `in_units` and `out_units` to be instances of "
                f"the Unit class. Instead encountered in_units: {in_units} and "
                f"out_units: {out_units}."
            )

    def _point_estimate(self):
        """Return a single point estimate in arbitrary units."""
        raise NotImplementedError()

    def point_estimate(self, units: Unit = Unit.ARBITRARY):
        """Return a single point estimate in the specified units."""
        return self._convert(self._point_estimate(), Unit.ARBITRARY, units)


class PointOutput(ModelOutput):
    """
    Class representing a batch of point estimates.
    """

    N_OUTPUTS_EXPECTED = 2

    config_name = "POINT"

    def _point_estimate(self):
        # return torch.clamp(self.raw_output, -1, 1)
        return self.raw_output


class ProbabilisticOutput(ModelOutput):
    """
    Base class packaging the raw Tensor output from a deep neural network that
    parameterizes a (batch of) distributions with its unit and choice of
    parameterization.
    """

    computes_calibration = True

    def _log_p(self, x: torch.Tensor) -> torch.Tensor:
        """
        Return log p(x) under the distribution p parameterized by the model
        output, where x is an array-like with expected shape (..., 2). Importantly,
        x is expected to be provided in arbitrary units (i.e. on the square [-1, 1]^2).
        """
        raise NotImplementedError

    def log_p(self, x: torch.Tensor, units: Unit) -> torch.Tensor:
        """
        Return log p(x) for x provided in units `units`.
        """
        # convert coordinate grid to arbitrary units
        x = x.to(self.device)
        x_arbitrary = self._convert(x, units, Unit.ARBITRARY)
        log_prob = self._log_p(x_arbitrary)
        # scale probability accordingly
        # jacobian determinant of the affine transform from [-1, 1]^2 \to
        # [0,xdim] x [0, ydim] is (scale_factor / 2)^ndim, so we divide the density by
        # that.
        scale_factor = (self.arena_dims[units].max() / 2) ** self.ndim
        # equivalently, subtract the log of the scale factor from log_prob.
        return log_prob - torch.log(scale_factor)

    def pmf(
        self, coordinate_grid: torch.Tensor, units: Unit, temperature: float = 1.0
    ) -> torch.Tensor:
        """
        Calculate p(x) at each point on the coordinate grid for the
        distribution p parameterized by this model output instance, in a
        vectorized and numerically stable way. Normalizes to sum to 1.

        Optionally, provide a `temperature` parameter to adjust the entropy of
        the distribution. Higher temperatures shift the distribution towards uniform,
        lower temperatures towards a point mass at the mode.
        """
        probs = torch.exp(
            self.log_p(coordinate_grid[..., None, :], units=units) / temperature
        )
        # normalize to 1
        return probs / probs.sum()


class BaseDistributionOutput(ProbabilisticOutput):
    """
    Utility subclass expressing that the model output
    parameterizes a single distribution rather than a mixture.
    """

    N_OUTPUTS_EXPECTED: int


class UniformOutput(BaseDistributionOutput):
    N_OUTPUTS_EXPECTED = 0
    config_name = "UNIFORM"

    def __init__(
        self,
        raw_output: None,
        arena_dims: torch.Tensor,
        arena_dim_units: Union[str, Unit],
        batch_size: int,
        device: str,
    ):
        """
        Inits UniformOutput.

        Args:
            raw_output: expected to be None. the argument is kept for consistency
                with other init methods, but it's ignored even if provided.
            arena_dims: array-like (x, y) storing arena dimensions in either CM or MM
            arena_dim_units: str or Unit indicating in which unit the arena dimensions
                are provided.
            batch_size: batch size. usually inferred from model output shape,
                but since no output is expected or considered here, it must be
                provided separately.
            device: device on which outputs should be stored. usually inferred as well.
        """
        super().__init__(raw_output, arena_dims, arena_dim_units)
        self.batch_size = batch_size
        self.device = device

    def _point_estimate(self):
        print(
            "Warning: method `_point_estimate` was called on a UniformOutput object. "
            "While supported for consistency, this method is almost nonsensical and "
            "should not be used in general.",
            file=sys.stderr,
        )
        return torch.zeros(self.batch_size, self.ndim, device=self.device)

    def _log_p(self, x: torch.Tensor) -> torch.Tensor:
        if x.shape[-2] != self.batch_size:
            raise ValueError(
                f"Incorrect shape for input! Since batch size is {self.batch_size}, "
                f"expected second-to-last dim of input `x` to have the same shape. Instead "
                f"found shape {x.shape}."
            )
        output_shape = x.shape[:-1]
        # Uniform distribution over [-ad[0]/ad.max, ad[0]/ad.max] x [-ad[1]/ad.max, ad[1]/ad.max] x ...
        # Volume of the arena is (2*ad[0] / ad.max) * (2*ad[1] / ad.max) * ... = 2^ndim * prod(ad) / ad.max^ndim
        volume = (
            torch.prod(self.arena_dims[Unit.MM])
            / (self.arena_dims[Unit.MM].max() ** self.ndim)
            * (2**self.ndim)
        )

        return torch.full(output_shape, -torch.log(volume), device=x.device)


class MDNOutput(ProbabilisticOutput):
    """
    Class storing the output of a mixture density network, which
    assumes a mixture response and parameterizes both the constituent
    distributions of the mixture as well as the weights by which the constituent
    distributions should be combined.
    """

    config_name = "MDN"

    def __init__(
        self,
        raw_output: torch.Tensor,
        arena_dims: torch.Tensor,
        arena_dim_units: Union[str, Unit],
        constituent_response_types: List[type[BaseDistributionOutput]],
        constituent_extra_kwargs: Optional[dict[int, dict]] = None,
    ):
        """
        Inits MDNOutput.

        Args:
            raw_output: batched tensor output from a VocalocatorArchitecture model
            arena_dims: array-like (x, y) storing arena dimensions in either CM or MM
            arena_dim_units: str or Unit indicating which unit the arena dimensions are provided in
            constituent_response_types: a list of class names indicating the distributions
                that are assumed to be components of a mixture response, with mixture weights specified
                on a per-stimulus basis by the model.
            constituent_extra_kwargs: an OPTIONAL extra dictionary mapping integer indices
                of the list `constituent_response_types` to keyword argument dictionaries
                that should be passed to the constituent distributions' constructors.
        """
        super().__init__(raw_output, arena_dims, arena_dim_units)

        # sum the number of parameters each constituent response distribution expects
        total_parameters_for_constituents = sum(
            t.N_OUTPUTS_EXPECTED for t in constituent_response_types
        )
        # then add on the number of mixing weights expected (R for R
        # the number of response distributions)
        num_responses = len(constituent_response_types)
        n_parameters_expected = total_parameters_for_constituents + num_responses

        if self.raw_output.shape[-1] != n_parameters_expected:
            raise ValueError(
                f"Given constituent response types {constituent_response_types}, expected "
                f"to recieve {n_parameters_expected} parameters per stimulus, with "
                f"{total_parameters_for_constituents} params to specify the constituent "
                f"distributions, and {num_responses} params to specify the mixing weights. "
                f"Instead encountered {raw_output.shape[-1]}, with raw_output shape "
                f"of {raw_output.shape}."
            )

        # we don't set `constituent_extra_kwargs` to the empty dict by default
        # because that's dangerous and stateful in python.
        if not constituent_extra_kwargs:
            constituent_extra_kwargs = {}

        curr_idx = 0
        self.responses = []
        for i, response_type in enumerate(constituent_response_types):
            n_params = response_type.N_OUTPUTS_EXPECTED
            next_idx = curr_idx + n_params

            additional_kwargs = constituent_extra_kwargs.get(i, {})

            if response_type == UniformOutput:
                additional_kwargs["batch_size"] = self.batch_size
                additional_kwargs["device"] = self.device

            response = response_type(
                raw_output[:, curr_idx:next_idx],
                arena_dims,
                arena_dim_units,
                **additional_kwargs,
            )
            self.responses.append(response)
            curr_idx = next_idx

        # the parameters representing the weights are expected to be
        # passes as unnormalized logits. call these w_ij for i the batch index
        # and j the logit for response distribution j.
        unnormalized_logits = raw_output[:, curr_idx:]
        # the actual weights theta_i are given by
        # theta_ij = exp(w_ij) / [ \sum_k exp(w_ik) ]
        # but we're ultimately going to work on log scale, so
        # instead compute log theta)ij = w_ij - log(sum_k exp(w_ik))
        normalizer = torch.logsumexp(unnormalized_logits, -1, keepdim=True)
        self.log_weights = unnormalized_logits - normalizer
        self.ndim = self.responses[0].ndim

    def _log_p(self, x: torch.Tensor) -> torch.Tensor:
        # with the log weights log(theta_ij) defined in the init
        # method, we can now compute the mixture density evaluated at a
        # certain (collection of) point(s)
        # stacks R tensors each of shape (..., self.batch_size)
        # into one big tensor of shape (..., self.batch_size, R)
        # this is for compatibility with self.log_weights, which has
        # shape (..., self.batch_size)
        individual_log_probs = torch.stack([r._log_p(x) for r in self.responses], -1)
        # the result is given by p_i = \sum_j theta_ij p_ij
        # for p_i the mixture density for batch element i,
        # and p_ij the probability assigned from response j
        # calculating this using only the log probs and log weights,
        # the formula becomes
        # log p_i = log \sum_j exp( log theta_ij + log p_ij)

        # tensor shape wise, this operation reduces along the last dimension
        # across the different responses as expected
        #     (..., self.batch_size, R) --> (..., self.batch_size)
        return torch.logsumexp(self.log_weights + individual_log_probs, -1)

    def _point_estimate(self):
        # average the component distribution's point estimates
        # throwing out any uniform distributions because a point estimate
        # from one is sort of nonsensical
        to_include = []
        for response in self.responses:
            if not isinstance(response, UniformOutput):
                to_include.append(response.point_estimate())

        point_estimates = torch.stack(to_include, dim=1)  # (self.batch_size, R, 2)
        weights = torch.exp(self.log_weights)  # (self.batch_size, R)
        reweighted_estimates = (
            point_estimates * weights[..., None]
        )  # (self.batch_size, R, 2)

        return reweighted_estimates.sum(dim=1)  # (self.batch_size, 2)

    @property
    def computes_calibration(self):
        return all(r.computes_calibration for r in self.responses)


class EnsembleOutput(ProbabilisticOutput):
    """
    Class storing the output of an ensemble of separate models.
    """

    config_name = "ENSEMBLE"

    def __init__(
        self,
        raw_output: List[ProbabilisticOutput],
        arena_dims: torch.Tensor,
        arena_dim_units: Union[str, Unit],
    ):
        """
        Initialize an object representing a mixture density created by
        averaging multiple separate probabilistic outputs, oftentimes from
        separate models.
        """
        super().__init__(raw_output, arena_dims, arena_dim_units)
        # make sure all constituent distributions have the same batch size
        self.distributions = raw_output
        self.n_distributions = len(self.raw_output)
        self.batch_size = self.raw_output[0].batch_size

        for output in self.raw_output:
            if output.batch_size != self.batch_size:
                raise ValueError(
                    "Not all input model output objects have the same batch size! "
                    f"The encountered batch sizes are: {[o.batch_size for o in self.raw_output]}."
                )

        self.ndim = self.distributions[0].ndim

    def _log_p(self, x: torch.Tensor) -> torch.Tensor:
        # output probability should be the average of the individual
        # distribution's probabilities. say i is batch index and p_ij is
        # probability from distribution j on batch element i
        # log p_i = log (1/N \sum_j p_ij) = log(1/N) + logsumexp(p_i1, ..., p_iN)
        # get each p_ij
        individual_log_probs = torch.stack(
            [r._log_p(x) for r in self.distributions], -1
        )
        # shape (..., self.batch_size, self.n_distributions) |--> (..., self.batch_size)
        unnormalized = torch.logsumexp(individual_log_probs, -1)
        normalizer = -torch.log(torch.tensor(self.n_distributions, device=x.device))
        return normalizer + unnormalized

    def _point_estimate(self):
        # average the component distribution's point estimates
        # throwing out any uniform distributions because a point estimate
        # from one is sort of nonsensical
        to_include = []
        for distr in self.distributions:
            if not isinstance(distr, UniformOutput):
                to_include.append(distr.point_estimate())
        return sum(to_include) / len(to_include)

    @property
    def computes_calibration(self):
        return all(r.computes_calibration for r in self.distributions)



from typing import Optional

import torch
from torch import nn
from torchaudio.functional import convolve

from vocalocator.architectures.base import VocalocatorArchitecture
from vocalocator.outputs import ModelOutputFactory


class WavenetBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        conv_channels: int,
        kernel_size: int,
        dilation: int,
        bias=True,
        *,
        injection_channels: Optional[int] = None,
    ):
        super().__init__()
        self.injection_linear = None
        if injection_channels is not None:
            self.injection_linear = nn.Linear(injection_channels, in_channels)

        self.conv = nn.Conv1d(
            in_channels=in_channels,
            out_channels=conv_channels * 2,
            kernel_size=kernel_size,
            stride=1,
            dilation=dilation,
            bias=bias,
            padding="same",
        )
        self.one_by_one = nn.Conv1d(
            in_channels=conv_channels,
            out_channels=in_channels,
            kernel_size=1,
            stride=1,
            dilation=1,
            bias=bias,
        )

    def forward(self, x: torch.Tensor, injection: Optional[torch.Tensor] = None):
        if self.injection_linear is not None:
            x = x + self.injection_linear(injection)[:, :, None]
        conv_out = self.conv(x)
        tanh, sigmoid = conv_out.chunk(2, dim=-2)
        tanh = torch.tanh(tanh) * 0.95 + 0.05 * tanh
        sigmoid = torch.sigmoid(sigmoid)
        activation = tanh * sigmoid
        one_by_one_output = self.one_by_one(activation)
        return one_by_one_output + x, one_by_one_output


def gradsafe_sum(x_list: list[torch.Tensor]) -> torch.Tensor:
    result = x_list[0]
    for x in x_list[1:]:
        result = result + x
    return result


class Wavenet(VocalocatorArchitecture):
    defaults = {
        "NUM_BLOCKS": 10,
        "CONV_CHANNELS": 64,
        "KERNEL_SIZE": 7,
        "DILATION": 3,
        "XCORR_PAIRS": None,
        "XCORR_LENGTH": 256,
        "XCORR_HIDDEN": 512,
    }

    def __init__(self, config: dict, output_factory: ModelOutputFactory):
        super().__init__(config, output_factory)

        N = config["DATA"]["NUM_MICROPHONES"]

        # Obtains model-specific parameters from the config file and fills in missing entries with defaults
        model_config = Wavenet.defaults.copy()
        model_config.update(config.get("MODEL_PARAMS", {}))
        config["MODEL_PARAMS"] = (
            model_config  # Save the parameters used in this run for backward compatibility
        )

        self.xcorr_mlp = None
        if model_config["XCORR_PAIRS"] is not None:
            self.xcorr_pairs = model_config["XCORR_PAIRS"]
            self.xcorr_length = model_config["XCORR_LENGTH"]
            xcorr_hidden = model_config["XCORR_HIDDEN"]

            self.xcorr_mlp = nn.Sequential(
                nn.BatchNorm1d(len(self.xcorr_pairs) * model_config["XCORR_LENGTH"]),
                nn.Linear(
                    len(self.xcorr_pairs) * model_config["XCORR_LENGTH"], xcorr_hidden
                ),
                nn.ReLU(),
                nn.Linear(xcorr_hidden, xcorr_hidden),
                nn.ReLU(),
            )

        self.blocks = nn.ModuleList(
            [
                WavenetBlock(
                    in_channels=model_config["CONV_CHANNELS"],
                    conv_channels=model_config["CONV_CHANNELS"],
                    kernel_size=model_config["KERNEL_SIZE"],
                    dilation=i % 4 + 1,
                    injection_channels=None if self.xcorr_mlp is None else xcorr_hidden,
                )
                for i in range(model_config["NUM_BLOCKS"])
            ]
        )
        self.initial_conv = nn.Conv1d(
            in_channels=N,
            out_channels=model_config["CONV_CHANNELS"],
            kernel_size=model_config["KERNEL_SIZE"],
            stride=1,
            dilation=1,
            padding="same",
        )

        self.n_outputs: int
        self.dense = nn.Sequential(
            nn.Linear(model_config["CONV_CHANNELS"], 512),
            nn.ReLU(),
            nn.Linear(512, self.n_outputs),
        )

    def make_xcorrs(self, x: torch.Tensor) -> torch.Tensor:
        """Computes cross-correlations between the configured channel pairs

        Args:
            x: Input tensor of shape (batch, channels, time)
        """
        xcorrs = []
        x_len = x.shape[-1]
        xcorr_len = self.xcorr_length
        centered_x = x[..., x_len // 2 - xcorr_len // 2 : x_len // 2 + xcorr_len // 2]
        for i, j in self.config["MODEL_PARAMS"]["XCORR_PAIRS"]:
            a = centered_x[:, i]
            # Use flip to compute cross-correlation instead of convolution
            b = torch.flip(centered_x[:, j], [-1])
            xcorr = convolve(a, b, mode="same")
            xcorrs.append(xcorr)

        return torch.cat(xcorrs, dim=-1)

    def _forward(self, x):
        x = torch.einsum("btc->bct", x)  # transpose
        xcorrs = None
        if self.xcorr_mlp is not None:
            xcorrs = self.make_xcorrs(x)
            xcorrs = self.xcorr_mlp(xcorrs)

        output = self.initial_conv(x)
        one_by_ones = []
        for block in self.blocks:
            output, obo = block(output, xcorrs)
            one_by_ones.append(obo)
        # Mean over time dim
        obo_sum = gradsafe_sum(one_by_ones).mean(dim=-1)
        output = self.dense(obo_sum)
        return output

    def clip_grads(self):
        nn.utils.clip_grad_norm_(self.parameters(), 1.0, error_if_nonfinite=True)



from .base import VocalocatorArchitecture
from .ensemble import VocalocatorEnsemble
from .simplenet import VocalocatorSimpleNetwork
from .simplenet_crosscorr import CorrSimpleNetwork
from .wavenet_arch import Wavenet

__all__ = [
    VocalocatorSimpleNetwork,
    VocalocatorEnsemble,
    CorrSimpleNetwork,
    Wavenet,
]



"""
Helper script to update parameters in a JSON file with new values from another.
"""

import argparse
from sys import stderr

try:
    # Attempt to use json5 if available
    import pyjson5 as json
except ImportError:
    print("Warning: json5 not available, falling back to json.", file=stderr)
    import json


def read_json(path) -> dict:
    with open(path) as f:
        return json.load(f)


def update_json(original: dict, changes: dict, outpath: str):
    # apply the changes and save to json
    original.update(changes)
    with open(outpath, "w") as f:
        json.dump(original, f)


if __name__ == "__main__":
    p = argparse.ArgumentParser(
        description="Update parameters in a JSON file with new values from another."
    )
    p.add_argument("original", type=str)
    p.add_argument("changes", type=str)
    p.add_argument("outpath", type=str)

    args = p.parse_args()

    original = read_json(args.original)
    changes = read_json(args.changes)

    update_json(original, changes, args.outpath)



from itertools import combinations
from math import comb

import torch
from torch import nn

from vocalocator.architectures.base import VocalocatorArchitecture
from vocalocator.outputs import ModelOutputFactory

from .simplenet import VocalocatorSimpleLayer


class SkipConnection(torch.nn.Module):
    def __init__(self, submodule: nn.Module):
        super(SkipConnection, self).__init__()
        self.submodule = submodule

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.submodule(x)


class CorrSimpleNetwork(VocalocatorArchitecture):
    defaults = {
        "USE_BATCH_NORM": True,
        "SHOULD_DOWNSAMPLE": [False, True] * 5,
        "CONV_FILTER_SIZES": [33] * 10,
        "CONV_NUM_CHANNELS": [16, 16, 32, 32, 64, 64, 128, 128, 256, 256],
        "CONV_DILATIONS": [1] * 10,
        "OUTPUT_COV": True,
        "REGULARIZE_COV": False,
        "CPS_NUM_LAYERS": 3,
        "CPS_HIDDEN_SIZE": 1024,
        "XCORR_LENGTH": 256,
    }

    def __init__(self, CONFIG, output_factory: ModelOutputFactory):
        super(CorrSimpleNetwork, self).__init__(CONFIG, output_factory)

        N = CONFIG["DATA"]["NUM_MICROPHONES"]

        # Obtains model-specific parameters from the config file and fills in missing entries with defaults
        model_config = CorrSimpleNetwork.defaults.copy()
        model_config.update(CONFIG.get("MODEL_PARAMS", {}))
        CONFIG["MODEL_PARAMS"] = (
            model_config  # Save the parameters used in this run for backward compatibility
        )

        self.xcorr_length = model_config["XCORR_LENGTH"]
        should_downsample = model_config["SHOULD_DOWNSAMPLE"]
        self.n_channels = model_config["CONV_NUM_CHANNELS"]
        filter_sizes = model_config["CONV_FILTER_SIZES"]
        dilations = model_config["CONV_DILATIONS"]

        min_len = min(
            len(self.n_channels),
            len(filter_sizes),
            len(should_downsample),
            len(dilations),
        )
        self.n_channels = self.n_channels[:min_len]
        filter_sizes = filter_sizes[:min_len]
        should_downsample = should_downsample[:min_len]
        dilations = dilations[:min_len]

        use_batch_norm = model_config["USE_BATCH_NORM"]
        self.n_channels.insert(0, N)

        convolutions = [
            VocalocatorSimpleLayer(
                in_channels,
                out_channels,
                filter_size,
                downsample=downsample,
                dilation=dilation,
                use_bn=use_batch_norm,
            )
            for in_channels, out_channels, filter_size, downsample, dilation in zip(
                self.n_channels[:-1],
                self.n_channels[1:],
                filter_sizes,
                should_downsample,
                dilations,
            )
        ]
        self.conv_layers = torch.nn.Sequential(*convolutions)

        self.final_pooling = nn.AdaptiveAvgPool1d(1)

        # layers for the cps branch of the network:
        cps_initial_channels = (
            comb(N, 2) * self.xcorr_length
        )  # Number of microphone pairs
        cps_num_layers = model_config["CPS_NUM_LAYERS"]
        cps_hidden_size = model_config["CPS_HIDDEN_SIZE"]

        # Add a batch norm to handle scaling the cps
        cps_network = [
            nn.BatchNorm1d(cps_initial_channels),
            nn.Linear(cps_initial_channels, cps_hidden_size),
            nn.ReLU(),
        ]
        for _ in range(cps_num_layers - 1):
            cps_network.append(
                SkipConnection(
                    nn.Sequential(
                        nn.Linear(cps_hidden_size, cps_hidden_size),
                        nn.ReLU(),
                    )
                )
            )
        self.cps_network = nn.Sequential(*cps_network)

        if not isinstance(self.n_outputs, int):
            raise ValueError(
                "Number of parameters to output is undefined! Maybe check the model configuration and ModelOutputFactory object?"
            )
        self.coord_readout = nn.Sequential(
            nn.Linear(
                self.n_channels[-1] + cps_hidden_size,
                self.n_channels[-1] + cps_hidden_size,
            ),
            nn.ReLU(),
            nn.Linear(self.n_channels[-1] + cps_hidden_size, self.n_outputs),
        )

    def make_cps(self, audio: torch.Tensor):
        # audio shape: (batch, n_samples, n_channels)
        # Generates the cross power spectrum of the audio channels in a pairwise fasion

        output_length = self.xcorr_length

        seq_len, n_channels = audio.shape[-2:]

        # Get all pairwise combinations of channels
        channel_pairs = list(combinations(range(n_channels), 2))
        channel_pairs = torch.tensor(channel_pairs)  # for 4 mics, shape is (6, 2)

        rfft_audio = torch.fft.rfft(audio, dim=-2)  # (batch, n_rfreq, n_channels)

        lhs = rfft_audio[:, :, channel_pairs[:, 0]]
        rhs = rfft_audio[:, :, channel_pairs[:, 1]]
        f_xcorr = lhs * rhs.conj()  # both complex, shape is (batch, n_rfreq, n_pairs)
        cps = torch.fft.irfft(f_xcorr, dim=-2)
        # shape is (batch, n_samples, n_pairs)

        # Take the central 256 samples
        center = cps.shape[-2] // 2
        width = output_length
        low_end = center - width // 2
        high_end = center + width // 2
        cps = cps[..., low_end:high_end, :]
        # shape is (batch, 256, n_pairs)

        return cps.flatten(start_dim=1)

    def _forward(self, x: torch.Tensor) -> torch.Tensor:
        audio = x
        cps = self.make_cps(audio)

        audio = audio.transpose(
            -1, -2
        )  # (batch, seq_len, channels) -> (batch, channels, seq_len) needed by conv1d
        h1 = self.conv_layers(audio)
        h2 = torch.squeeze(self.final_pooling(h1), dim=-1)

        cps_branch = self.cps_network(cps)
        h2 = torch.cat((h2, cps_branch), dim=-1)

        coords = self.coord_readout(h2)
        return coords

    def clip_grads(self):
        nn.utils.clip_grad_norm_(self.parameters(), 1.0, error_if_nonfinite=True)



"""
Functions used in calculating the calibration of a model outputting mean + cov.
"""

import logging
from typing import Tuple, Union

import numpy as np
import torch

from vocalocator.outputs.base import ProbabilisticOutput, Unit
from vocalocator.util import assign_to_bin_2d, digitize, make_xy_grids

logger = logging.getLogger(__name__)


def min_mass_containing_location(
    pmf: np.ndarray, location: np.ndarray, xgrid: np.ndarray, ygrid: np.ndarray
):
    """
    Return the minimum probability mass containing the true location. This
    function is used to calculate the 95% confidence set, as well as calibration
    curves.
    """
    # maps: (n_y_bins, n_x_bins)
    # locations: (NUM_SAMPLES, 2)
    # coord_bins: (n_y_bins + 1, n_x_bins + 1, 2)  ( output of meshgrid then dstack )
    n_y_bins, n_x_bins = pmf.shape

    # first verify that xgrid and ygrid have correct shapes
    # since xgrid and ygrid should represent the EDGES of the gridpoints at
    # which the pmf is evaluated
    expected_grid_shape = (n_y_bins + 1, n_x_bins + 1)
    if xgrid.shape != expected_grid_shape or ygrid.shape != expected_grid_shape:
        raise ValueError(
            f"Expected `xgrid` and `ygrid` to have shape {expected_grid_shape}, "
            f"since the pmf has shape {pmf}. "
            f"Instead, encountered `xgrid` shape {xgrid.shape} and "
            f"`ygrid` shape {ygrid.shape}"
        )

    # flatten the pmf
    flattened = pmf.flatten()
    # argsort in descending order
    argsorted = flattened.argsort()[::-1]
    # assign the true location to a coordinate bin
    # reshape loc to a (1, 2) array so the vectorized function
    # assign_to_bin_2d still works
    location = location.reshape(1, 2)
    loc_idx = assign_to_bin_2d(location, xgrid, ygrid)
    # bin number for first interval containing location
    bin_idx = (argsorted == loc_idx).argmax()
    # distribution with values at indices above bin_idxs zeroed out
    # x_idx = [
    # [0, 1, 2, 3, ...],
    # [0, 1, 2, 3, ...]
    # ]
    num_bins = pmf.size
    sorted_maps = flattened[argsorted]
    s = np.where(np.arange(num_bins) > bin_idx, 0, sorted_maps).sum()
    if (s < 0).any():
        logger.warning(f"negative min masses containing locations found: {s}")
    # clip values to remove any floating point errors
    # where the sum is greater than 1
    return s.clip(0, 1)


def calculate_confidence_set(pmf: np.ndarray, threshold: float):
    """
    Return the region around the mean to which the model assigns
    `threshhold` proprtion of the probability mass.
    """
    flattened = pmf.flatten()
    sorted_bins = flattened.argsort()[::-1]

    # select bins in descending order by mass until we hit `threshhold`
    sorted_masses = flattened[sorted_bins]
    total_mass_by_bins = sorted_masses.cumsum()

    bins_in_confidence_set = total_mass_by_bins <= threshold

    # reorder these values from their descended sorting pattern
    # back to the original places of the bins
    restore_bin_order = sorted_bins.argsort()
    confidence_set = bins_in_confidence_set[restore_bin_order]

    confidence_set = confidence_set.reshape(pmf.shape)

    return confidence_set


def confidence_set_stats(
    pmf: np.ndarray,
    threshhold: float,
    arena_dims: Union[Tuple[float, float], np.ndarray],
    true_location: np.ndarray,
) -> tuple[np.ndarray, float, bool, float]:
    """
    Given a pmf, compute and return the `threshhold` confidence set, along with
    relevant info like its area, whether the true location is in the set, etc.
    """
    # find the mean of each pmf
    center_xgrid, center_ygrid = make_xy_grids(
        arena_dims, shape=pmf.shape, return_center_pts=True
    )

    # pmf shape: (n_y_pts, n_x_pts)
    # marginal distributions over x: pmf.sum(axis=0)
    p_X = pmf.sum(axis=0)
    x_coords = (p_X * center_xgrid[0]).sum()
    # marginal distribution over y: pmf.sum(axis=1)
    p_Y = pmf.sum(axis=1)
    y_coords = (p_Y * center_ygrid[:, 0]).sum()
    means = np.column_stack((x_coords, y_coords))

    # get the confidence sets
    confidence_set = calculate_confidence_set(pmf, threshhold)

    # was the true location in the confidence sets?
    edge_xgrid, edge_ygrid = make_xy_grids(arena_dims, shape=np.array(pmf.shape) + 1)
    loc_bin = assign_to_bin_2d(true_location.reshape(1, 2), edge_xgrid, edge_ygrid)
    (y_idx, x_idx) = np.unravel_index(loc_bin, shape=pmf.shape)
    loc_in_confidence_set = bool(confidence_set[y_idx, x_idx])

    # find the area of each set
    arena_area = arena_dims[0] * arena_dims[1]
    num_bins_per_set = confidence_set.sum()
    area = float((num_bins_per_set / confidence_set.size) * arena_area)

    # find the distances from the furthest point in the calibration set
    # to the mean
    center_coord_grid = np.dstack((center_xgrid, center_ygrid))
    distances = np.linalg.norm(center_coord_grid - means[None], axis=-1)

    distances_in_set = np.where(confidence_set, distances, 0)
    dist_to_furthest_point = float(distances_in_set.max())

    return (confidence_set, area, loc_in_confidence_set, dist_to_furthest_point)


class CalibrationAccumulator:
    """
    Helper class to simplify the calculation of calibration,
    with support for multiple output types and applying various
    smoothing methods to each output type.
    """

    def __init__(
        self,
        arena_dims: Union[Tuple[float, float], np.ndarray],
        n_calibration_bins: int = 10,
        confidence_set_threshold: float = 0.95,
    ):
        """
        Initialize a CalibrationAccumulator object, which is a helper class
        to simplify the calculation of calibration in an online manner (iterating
        through or asynchronously going through a dataset).
        """
        self.n_calibration_bins = n_calibration_bins

        self.arena_dims = arena_dims  # expected to be in MM.

        # initialize the internal mass counts tracking arrays
        self.mass_counts = np.zeros(n_calibration_bins)

        # and the outputs for confidence set calculation
        self.confidence_set_threshold = confidence_set_threshold

        self.confidence_sets = []
        self.confidence_set_areas = []
        self.location_in_confidence_set = []
        self.distances_to_furthest_point = []

    def calculate_step(
        self,
        model_output: ProbabilisticOutput,
        true_location: np.ndarray,
        temperature: float = 1.0,
    ):
        """
        Perform one step of the calibration process on `model_output`.

        Essentially, this function calculates the probability assigned to the
        smallest region in the xy plane containing the true location. These
        regions are defined by progressively taking the location bins to which
        the model assigns the highest probability mass.

        Optionally, temperature scale the model output before calculating
        calibration, as per "On Calibration of Modern Neural Networks" (Guo, 2017).
        """
        if not isinstance(model_output, ProbabilisticOutput):
            raise ValueError(
                "Model output passed to calibration expected to be a subclass of "
                f"`ProbabilisticOutput`! Instead encountered object of type: {type(model_output)}."
            )
        # NOTE: true location expected in MM

        # get the pmf
        coords = self._make_coord_array()
        # add a batch dimension to match expected shape from `ProbabilisticOutput.pmf`
        coords = np.expand_dims(coords, -2)
        pmf = (
            model_output.pmf(torch.tensor(coords), Unit.MM, temperature=temperature)
            .cpu()
            .numpy()
        )
        # get rid of the extra batch dimension
        pmf = pmf.squeeze()

        # calculate the confidence set for this prediction
        # and store useful stats about it
        (
            confidence_set,
            area,
            loc_in_confidence_set,
            dist_to_furthest_point,
        ) = confidence_set_stats(
            pmf, self.confidence_set_threshold, self.arena_dims, true_location
        )

        self.confidence_sets.append(confidence_set)
        self.confidence_set_areas.append(area)
        self.location_in_confidence_set.append(loc_in_confidence_set)
        self.distances_to_furthest_point.append(dist_to_furthest_point)

        # get our x and ygrids to match the shape of the pmfs
        # since the grids track the edge points, we should have
        # one more point in each coordinate direction.
        grid_shape = np.array(pmf.shape) + 1
        xgrid, ygrid = make_xy_grids(self.arena_dims[:2], shape=grid_shape)

        # reshape location to (1, 2) if necessary
        # we do this so the repeat function works out correctly
        if true_location.shape != (1, 2):
            true_location = true_location.reshape((1, 2))

        # perform the calibration calculation step
        mass = min_mass_containing_location(pmf, true_location, xgrid, ygrid)

        # transform to the bin in [0, 1] to which each value corresponds,
        # essentially iteratively building a histogram with each step
        bins = np.arange(self.n_calibration_bins + 1) / self.n_calibration_bins
        bin_idx = digitize(mass, bins)

        # update mass counts array
        self.mass_counts[bin_idx] += 1

    def results(self):
        """
        Calculate calibration curves and error from the collected
        results of all the calibration steps.
        """
        calibration_curve = self.mass_counts.cumsum() / self.mass_counts.sum()
        # prepend a zero to the calibration curve so each value represents
        # the proportion of values whose min_mass is less than the bin
        calibration_curve = np.pad(calibration_curve, (1, 0), "constant")

        results = {}
        results["calibration_curve"] = calibration_curve

        results["confidence_sets"] = self.confidence_sets
        results["confidence_set_areas"] = self.confidence_set_areas
        results["location_in_confidence_set"] = self.location_in_confidence_set
        results["distances_to_furthest_point"] = self.distances_to_furthest_point

        return results

    def _make_coord_array(self):
        """
        Return a grid of evenly spaced points on the arena floor.
        """
        xdim = self.arena_dims[0]
        ydim = self.arena_dims[1]

        # change xgrid / ygrid size to preserve aspect ratio
        ratio = ydim / xdim
        desired_shape = (int(ratio * 100), 100)
        xgrid, ygrid = make_xy_grids(
            (xdim, ydim), shape=desired_shape, return_center_pts=True
        )
        coords = np.dstack((xgrid, ygrid)).astype(np.float32)
        return coords



from pathlib import Path
from typing import Literal, overload

import torch
from vocalocator.outputs import ModelOutput, ModelOutputFactory


class VocalocatorArchitecture(torch.nn.Module):
    defaults: dict

    def __init__(self, CONFIG, output_factory: ModelOutputFactory):
        super(VocalocatorArchitecture, self).__init__()

        self.config = CONFIG
        self.output_factory = output_factory
        self.n_outputs = self.output_factory.n_outputs_expected

    def load_weights(self, best_weights_path=None, use_final_weights: bool = False):
        """
        Load in the weights of the model, where use of `best_weights.pt` or
        `final_weights.pt` is specified by `use_final_weights`.
        """
        if best_weights_path is None:
            # try checking config for a weights path
            print("No weights path provided, checking model config.")
            try:
                best_weights_path = self.config["WEIGHTS_PATH"]
            except KeyError:
                raise ValueError(
                    "Couldn't load model weights! No `best_weights_path` kwarg provided "
                    'to `load_weights`, and could not find "WEIGHTS_PATH" in model config.'
                )

        weights_path = best_weights_path
        if use_final_weights:
            filename = Path(weights_path).name
            if "best" not in filename:
                raise ValueError(
                    "Since best weights path isn't of the form /.../best_weights.pt, "
                    "cannot locate a corresponding final weights file!"
                )
            else:
                filename = filename.replace("best", "final")
                weights_path = Path(weights_path).parent / filename

        device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.to(device)

        print(f"Loading weights from path {weights_path}.")
        weights = torch.load(weights_path, map_location=device)
        self.load_state_dict(weights)

    def _forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    # add overload for nice unbatched functionality
    @overload
    def forward(self, x: torch.Tensor, unbatched: Literal[False]) -> ModelOutput: ...

    @overload
    def forward(
        self, x: torch.Tensor, unbatched: Literal[True]
    ) -> list[ModelOutput]: ...

    def forward(self, x: torch.Tensor, unbatched: bool = False):
        """
        Run the model on input `x` and return an appropriate
        ModelOutput object, determined based on `self.output_factory`.
        """
        raw_outputs = self._forward(x)
        if unbatched:
            outputs = []
            for o in raw_outputs:
                # add batch dim back
                outputs.append(self.output_factory.create_output(o[None]))
            return outputs
        else:
            return self.output_factory.create_output(self._forward(x))



"""
Test the build model functions.
"""

import unittest

import torch
from constants import ENSEMBLE, ENSEMBLE_MISSING_COV, SIMPLENET_BASE, SIMPLENET_COV
from vocalocator.training.models import build_model


class TestOuptutShapes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # number of channels of input with four mics and passed-in cross correlations
        NUM_CHANNELS_MICS_AND_XCORR = 10
        cls.fake_input = torch.zeros(
            1, SIMPLENET_BASE["SAMPLE_LEN"], NUM_CHANNELS_MICS_AND_XCORR
        )

    def test_simplenet(self):
        # test loading the models
        cnn_no_cov, _ = build_model(SIMPLENET_BASE)
        cnn_cov, _ = build_model(SIMPLENET_COV)

        no_cov_expected_shape = (1, 2)
        cov_expected_shape = (1, 3, 2)

        self.assertEqual(cnn_no_cov(self.fake_input).shape, no_cov_expected_shape)
        self.assertEqual(cnn_cov(self.fake_input).shape, cov_expected_shape)

    def test_ensemble(self):
        ensemble, _ = build_model(ENSEMBLE)
        expected_shape = (1, len(ENSEMBLE["MODELS"]), 3, 2)
        self.assertEqual(ensemble(self.fake_input).shape, expected_shape)

        # test that the class throws an error if one of the models
        # doesn't have the 'OUTPUT_COV' flag
        with self.assertRaises(ValueError):
            build_model(ENSEMBLE_MISSING_COV)



"""
Assess covariance models on a dataset, tracking metrics like mean error,
area of the 95% confidence set for each prediction
and whether the true value was in that set, etc.
"""

import argparse
import json
import logging
from pathlib import Path
from typing import Union

import h5py
import numpy as np
import torch
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader
from tqdm import tqdm

from vocalocator.architectures.base import VocalocatorArchitecture
from vocalocator.architectures.ensemble import VocalocatorEnsemble
from vocalocator.calibration import CalibrationAccumulator
from vocalocator.outputs.base import ModelOutput, ProbabilisticOutput, Unit
from vocalocator.training.configs import build_config
from vocalocator.training.dataloaders import VocalizationDataset
from vocalocator.training.models import build_model
from vocalocator.util import make_xy_grids, subplots

logging.getLogger("matplotlib").setLevel(logging.WARNING)

FIRST_N_VOX_TO_PLOT = 16


def plot_results(f: h5py.File):
    """
    Create and save plots of calibration curve, error distributions, etc.
    """
    point_predictions = f["point_predictions"][:]
    scaled_locations = f["scaled_locations"][:].reshape(point_predictions.shape[0], -1)

    errs = np.linalg.norm(point_predictions - scaled_locations, axis=-1)
    if "calibration_curve" not in f.attrs:
        fig, err_ax = plt.subplots(1, 1, sharex=False, sharey=False)

        # convert errs to cm for readability
        errs_cm = errs / 10

        err_ax.hist(errs_cm, color="tab:blue")
        err_ax.set_xlabel("errors (cm)")
        err_ax.set_ylabel("counts")
        err_ax.set_title("error distribution")
        return fig, err_ax
    else:
        fig, axs = plt.subplots(1, 5, sharex=False, sharey=False, figsize=(20, 4))
        (err_ax, calib_ax, cset_area_ax, cset_radius_ax, dist_ax) = axs.flat

        # convert errs to cm for readability
        errs_cm = errs / 10

        err_ax.hist(errs_cm, color="tab:blue")
        err_ax.set_xlabel("errors (cm)")
        err_ax.set_ylabel("counts")
        err_ax.set_title("error distribution")

        calib_ax.plot(np.linspace(0, 1, 11), f.attrs["calibration_curve"][:], "bo")
        calib_ax.set_xlabel("probability assigned to region")
        calib_ax.set_ylabel("proportion of locations in the region")
        calib_ax.set_title("calibration curve")

        # convert confidence set areas to cm^2 for readability
        cset_areas_cm = f["confidence_set_areas"][:] / 100

        cset_area_ax.hist(cset_areas_cm, color="tab:blue")
        cset_area_ax.set_xlabel("confidence set area (cm^2)")
        cset_area_ax.set_ylabel("counts")
        cset_area_ax.set_title("confidence set area distribution")

        cset_radius_ax.plot(np.sqrt(cset_areas_cm), errs_cm, "bo")
        cset_radius_ax.set_xlabel("square root confidence set area (cm)")
        cset_radius_ax.set_ylabel("error (cm)")
        cset_radius_ax.set_title("sqrt confidence set area vs error")

        # convert distances to cm
        dist_ax.plot(f["distances_to_furthest_point"][:] / 10, errs_cm, "bo")
        dist_ax.set_xlabel("distance to furthest point in confidence set (cm)")
        dist_ax.set_ylabel("error (cm)")
        dist_ax.set_title("distance to furthest point vs error")

        return fig, axs


def assess_model(
    model: VocalocatorArchitecture,
    dataloader: DataLoader,
    outfile: Union[Path, str],
    arena_dims: Union[np.ndarray, tuple[float, float]],
    device: Union[str, torch.device] = "cuda:0",
    visualize: bool = False,
    temperature: float = 1.0,
    inference: bool = False,
):
    """
    Assess the provided model with uncertainty, storing model output as well as
    info like error, confidence sets, and a calibration curve in the h5 format
    at path `outfile`.

    Optionally, visualize confidence sets a few times throughout training.

    Args:
        model: instantiated VocalocatorArchitecture object
        dataloader: DataLoader object on which the model should be assessed
        outfile: path to an h5 file in which output should be saved
        arena_dims: arena dimensions, *in millimeters*.
        visualize: optional, indicates whether first few outputs should be plotted
        temperature: optional, adjusts entropy of probabilistic model outputs
    """
    outfile = Path(outfile)

    N = len(dataloader.dataset)

    with h5py.File(outfile, "w") as f:
        # Save the config data in the h5 file for future reference
        config_string = json.dumps(model.config)
        f.attrs["model_config"] = config_string

        scaled_locations_dataset = None

        if isinstance(model, VocalocatorEnsemble):
            raw_output_dataset = []
            for i, constituent in enumerate(model.models):
                raw_output_dataset.append(
                    f.create_dataset(
                        f"constituent_{i}_raw_output", shape=(N, constituent.n_outputs)
                    )
                )
        else:
            raw_output_dataset = f.create_dataset(
                "raw_model_output", shape=(N, model.n_outputs)
            )

        point_predictions = None

        ca = CalibrationAccumulator(arena_dims)

        model.eval()

        should_compute_calibration = False

        with torch.no_grad():
            idx = 0
            for sounds, locations in tqdm(dataloader):
                sounds = sounds.to(device)
                # If inference, locations will be None

                outputs: list[ModelOutput] = model(sounds, unbatched=True)
                for output, location in zip(outputs, locations):
                    # add batch dimension back
                    if isinstance(model, VocalocatorEnsemble):
                        for out_dataset, constituent_output in zip(
                            raw_output_dataset, output.raw_output
                        ):
                            out_dataset[idx] = (
                                constituent_output.raw_output.squeeze().cpu().numpy()
                            )
                    else:
                        raw_output_dataset[idx] = (
                            output.raw_output.squeeze().cpu().numpy()
                        )

                    point_est = output.point_estimate(units=Unit.MM).cpu().numpy()
                    if point_predictions is None:
                        point_predictions = f.create_dataset(
                            "point_predictions", shape=(N, *point_est.shape[1:])
                        )
                    point_predictions[idx] = point_est

                    # unscale location from [-1, 1] square to units in arena (in mm)
                    if location is not None:
                        scaled_location = (
                            output._convert(location[None], Unit.ARBITRARY, Unit.MM)
                            .cpu()
                            .numpy()
                        )
                        if scaled_locations_dataset is None:
                            scaled_locations_dataset = f.create_dataset(
                                "scaled_locations",
                                shape=(N, *scaled_location.shape[1:]),
                            )
                        scaled_locations_dataset[idx] = scaled_location

                        # other useful info
                        if output.computes_calibration and not inference:
                            should_compute_calibration = True
                            ca.calculate_step(
                                output, scaled_location, temperature=temperature
                            )

                            if visualize and idx == FIRST_N_VOX_TO_PLOT:
                                # plot the densities
                                visualize_dir = outfile.parent / "pmfs_visualized"
                                visualize_dir.mkdir(exist_ok=True, parents=True)
                                visualize_outfile = (
                                    visualize_dir / f"{outfile.stem}_visualized.png"
                                )

                                _, axs = subplots(FIRST_N_VOX_TO_PLOT)

                                for i, ax in enumerate(axs):
                                    ax.set_title(f"vocalization {i}")
                                    ax.plot(
                                        *point_predictions[i], "ro", label="predicted"
                                    )
                                    # add a green dot indicating the true location
                                    ax.plot(
                                        *scaled_locations_dataset[i], "go", label="true"
                                    )
                                    ax.set_aspect("equal", "box")

                                    if should_compute_calibration:
                                        set_to_plot = ca.confidence_sets[i]

                                        xgrid, ygrid = make_xy_grids(
                                            arena_dims,
                                            shape=set_to_plot.shape,
                                            return_center_pts=True,
                                        )
                                        ax.contourf(
                                            xgrid,
                                            ygrid,
                                            set_to_plot,
                                            label="95% confidence set",
                                        )
                                    ax.legend()

                                plt.savefig(visualize_outfile)
                                print(
                                    f"Model output visualized at file {visualize_outfile}"
                                )

                    # update number of vocalizations seen
                    idx += 1

        # Cannot compute calibration curve on inference data
        if should_compute_calibration and not inference:
            results = ca.results()
            f.attrs["calibration_curve"] = results["calibration_curve"]

            # array-like quantities outputted by the calibration accumulator
            OUTPUTS = (
                "confidence_sets",
                "confidence_set_areas",
                "location_in_confidence_set",
                "distances_to_furthest_point",
            )
            for output_name in OUTPUTS:
                f.create_dataset(output_name, data=results[output_name])

        if not inference:
            _, axs = plot_results(f)
            plt.tight_layout()
            plt.savefig(Path(outfile).parent / f"{Path(outfile).stem}.png")

    print(f"Model output saved to {outfile}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Used to specify model configuration via a JSON file.",
    )

    parser.add_argument(
        "--data",
        type=str,
        required=True,
        help="Path to an h5 file on which to assess the model",
    )

    parser.add_argument(
        "-o",
        "--outfile",
        type=str,
        required=True,
        help="Path at which to store results. Must be an h5 file.",
    )

    parser.add_argument(
        "--visualize",
        action="store_true",
        help="Include flag to plot confidence sets occasionally during assessment.",
    )

    parser.add_argument(
        "--use-final",
        action="store_true",
        help="Include flag use the FINAL model weights, not the best.",
    )

    parser.add_argument(
        "--temperature",
        type=float,
        required=False,
        default=1.0,
        help="Optional flag to apply temperature scaling to a model with probabilistic output.",
    )

    parser.add_argument(
        "--inference",
        action="store_true",
        help="Include flag to evaluate the model on datasets without ground truth.",
    )

    parser.add_argument(
        "--index",
        type=Path,
        required=False,
        help="Optional path to an index file to use for assessement. Should be a numpy int array of indices into the dataset.",
    )

    args = parser.parse_args()

    if not Path(args.config).exists():
        raise ValueError(
            f"Requested config JSON file could not be found: {args.config}"
        )

    config_data = build_config(args.config)

    model, _ = build_model(config_data)

    best_weights_path = config_data.get("WEIGHTS_PATH", None)
    model.load_weights(
        best_weights_path=best_weights_path, use_final_weights=args.use_final
    )

    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    if device == "cpu":
        config_data["GENERAL"]["DEVICE"] = "cpu"
    model = model.to(device)

    arena_dims = np.array(config_data["DATA"]["ARENA_DIMS"])
    arena_dims_units = config_data["DATA"].get("ARENA_DIMS_UNITS")
    sample_rate = config_data["DATA"]["SAMPLE_RATE"]
    crop_length = config_data["DATA"]["CROP_LENGTH"]
    normalize_data = config_data["DATA"].get("NORMALIZE_DATA", True)
    node_names = config_data["DATA"].get("NODES_TO_LOAD", None)
    vocalization_dir = config_data["DATA"].get(
        "VOCALIZATION_DIR",
        None,
    )

    # if provided in cm, convert to MM
    if arena_dims_units == "MM":
        pass
    elif arena_dims_units == "CM":
        arena_dims = np.array(arena_dims) * 10
    elif arena_dims_units == "M":
        arena_dims = np.array(arena_dims) * 1000
    else:
        raise ValueError(
            "ARENA_DIMS_UNITS must be one of 'MM,' 'CM,' or 'M' to specify the units of the arena dimensions."
        )

    index = None
    if args.index is not None:
        if not Path(args.index).exists():
            raise ValueError(f"Requested index file could not be found: {args.index}")
        index = np.load(args.index)

    dataset = VocalizationDataset(
        str(args.data),
        arena_dims=arena_dims,
        crop_length=crop_length,
        inference=args.inference,
        index=index,
        normalize_data=normalize_data,
        sample_rate=sample_rate,
        sample_vocalization_dir=vocalization_dir,
        nodes=node_names,
    )

    batch_size = config_data["DATA"]["BATCH_SIZE"]
    dataloader = DataLoader(
        dataset, batch_size=batch_size, shuffle=False, collate_fn=dataset.collate
    )

    # make the parent directories for the desired outfile if they don't exist
    parent = Path(args.outfile).parent
    parent.mkdir(exist_ok=True, parents=True)

    assess_model(
        model,
        dataloader,
        args.outfile,
        arena_dims,
        device=device,
        visualize=args.visualize,
        temperature=args.temperature,
        inference=args.inference,
    )



"""Loss functions for the DNNs."""

import logging

import torch
from torch.distributions import MultivariateNormal
from torch.nn import functional as F

from vocalocator.outputs.base import ModelOutput, ProbabilisticOutput

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# MARK: Loss functions


def squared_error(output: ModelOutput, target: torch.Tensor):
    """Mean squared error criterion."""
    return 2 * torch.mean(torch.square(output.point_estimate() - target))


def negative_log_likelihood(output: ProbabilisticOutput, target: torch.Tensor):
    """Negative log likelihood loss criterion."""
    return -1 * output._log_p(target)


# All loss functions are expected to accept the prediction as the first argument
# and the ground truth as the second argument
def se_loss_fn(pred: torch.Tensor, target: torch.Tensor):
    # Assumes the inputs have shape (B, 2), representing a batch of `B` 2-dimensional coordinates
    return 2 * torch.mean(torch.square(target - pred))  # Should be a scalar tensor


def map_se_loss_fn(pred: torch.Tensor, target: torch.Tensor):
    """MSE loss over a location map."""
    target = torch.flatten(target, start_dim=1)
    pred = torch.flatten(pred, start_dim=1)
    return torch.mean(torch.square(target - pred).mean(dim=1))


def wass_loss_fn(pred: torch.Tensor, target: torch.Tensor):
    """Calculates the earth mover's distance between the target and predicted locations.
    This is done by element-wise multiplying the prediction map by a (pre-computed) matrix
    of distances, in which each element contains the distance of that bin from the target
    location.
    This implementation is not symmetric because it scales the input 'pred' via a softmax.
    What is returned is not the true distance, but a numerically stable analogue that shares
    minima with the true distance w.r.t model parameters.
    """
    flat_pred = torch.flatten(pred, start_dim=1)
    flat_target = torch.flatten(target, start_dim=1)
    # Add 1 to target to make the smallest value 0
    error_map = F.log_softmax(flat_pred, dim=1) + torch.log(flat_target + 1)
    return torch.logsumexp(error_map, dim=1)


def gaussian_NLL(pred: torch.Tensor, target: torch.Tensor):
    """
    Negative log likelihood of the Gaussian parameterized by the model prediction.

    Assumes that the conditional distribution of a location y
    given an example x can be approximated by a multivariate gaussian

        p(y | x) = N(mu, Sigma)

    where mu(x) and Sigma(x) are models of the mean and covariance as a
    function of the audio.

    Expects pred to be a tensor of shape (B, 3, 2), where pred[i][0]
    represents the predicted mean for example i and the entries pred[i][1:]
    and the remaining entries represent a lower triangular matrix L such that
    L @ L.T = Sigma(x).
    """
    # make sure that preds has shape: (B, 3, 2) where B is the batch size
    if pred.ndim != 3:
        raise ValueError(
            "Expected `pred` to have shape (B, 3, 2) where B is the batch size, "
            f"but encountered shape: {pred.shape}"
        )

    y_hat = pred[:, 0]  # output, shape: (B, 2)
    L = pred[:, 1:]  # cholesky factor of covariance, shape: (B, 2, 2)

    # create the distribution corresponding to the outputted predictive
    # density
    multivariate_normal = MultivariateNormal(loc=y_hat, scale_tril=L)
    loss = -multivariate_normal.log_prob(target)

    return loss


def gaussian_NLL_half_normal_variances(pred: torch.Tensor, target: torch.Tensor):
    """
    Regularized version of the gaussian_NLL loss function. Specifically, this
    is the negative log posterior p(mu, Sigma | x) where we place half-Normal
    priors on the diagonal entries of the covariance matrix.

    Motivation: no prior belief about the skew direction of the model's
    confidence ellipses, so we should have a uniform prior over the correlation
    structure — reasonable choice is the LJK distribution with shape parameter
    eta = 1. Technically, we also implicitly put a uniform prior over the
    means, which are constrained to be on the unit square [-1, 1]^2. But since
    these are uniform, they don't affect the negative log posterior. The only
    thing that does is our half-Normal priors on the variances sigma^2_x and
    sigma^2_y — after taking the negative log, this results in just adding

        lambda_x sigma^2_x + lambda_y sigma^2_y

    to the output of the gaussian_NLL function, where lambda_i is half the
    squared scale paramater to the half-Normal. Since the outputs are scaled to the square
    [-1, 1]^2, picking prior variances to be 1 expresses the belief
    that about 68% of the time, the variance in the x and y direction should be
    less than half the arena size.
    """
    NLL_term = gaussian_NLL(pred, target)

    cholesky_cov = pred[:, 1:]
    cov = cholesky_cov @ cholesky_cov.swapaxes(-2, -1)

    variance_x = cov[:, 0, 0]
    variance_y = cov[:, 1, 1]

    return NLL_term + 0.5 * (variance_x + variance_y)


def gaussian_NLL_entropy_penalty(
    pred: torch.Tensor,
    target: torch.Tensor,
    penalty: float = 1.0,
):
    """
    Regularized version of gaussian_NLL, with an entropy penalty.
    """
    # make sure that preds has shape: (B, 3, 2) where B is the batch size
    if pred.ndim != 3:
        raise ValueError(
            "Expected `pred` to have shape (B, 3, 2) where B is the batch size, "
            f"but encountered shape: {pred.shape}"
        )

    y_hat = pred[:, 0]  # output, shape: (B, 2)
    L = pred[:, 1:]  # cholesky factor of covariance, shape: (B, 2, 2)

    # create the distribution corresponding to the outputted predictive
    # density
    multivariate_normal = MultivariateNormal(loc=y_hat, scale_tril=L)

    loss = -multivariate_normal.log_prob(target) + (
        penalty * multivariate_normal.entropy()
    )

    return loss



import torch
from numpy import sqrt
from torch import nn


class Identity(nn.Module):
    """Implements the identity function as a placeholder augmentation"""

    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


class PolarityInversion(nn.Module):
    """Inverts the polarity of the audio."""

    def __init__(self, p: float = 0.5):
        super().__init__()
        self.p = p

    def forward(self, x):
        if torch.rand(1) > self.p:
            return x
        return -x


class TimeMask(nn.Module):
    """Time masking data augmentation. Masks a random part of the audio across all channels.
    Mutates the input tensor in-place. A bit slow because it has to do a lot of indexing.
    """

    def __init__(
        self, min_mask_length: int = 0, max_mask_length: int = 0, prob: float = 0.5
    ):
        super().__init__()
        self.min_mask_length = min_mask_length
        self.max_mask_length = max_mask_length
        self.prob = prob

    def forward(self, x):
        if x.dim() <= 2:
            # unbatched input
            if torch.rand() > self.prob:
                return x
            num_samples, num_channels = x.shape
            mask_start = torch.randint(0, num_samples - self.max_mask_length, (1,))
            mask_end = mask_start + torch.randint(
                self.min_mask_length, self.max_mask_length, (1,)
            )
            x[mask_start:mask_end, :] = 0
            return x
        # Batched input
        bshape = x.shape[:-2]
        num_samples, num_channels = x.shape[-2:]
        x = x.reshape(-1, num_samples, num_channels)
        bsz = x.shape[0]

        prob_mask = torch.rand(bsz) < self.prob  # True where we should apply the mask
        prob_idx = torch.nonzero(prob_mask).squeeze(1)
        num_true = prob_idx.shape[0]
        mask_start = torch.randint(0, num_samples - self.max_mask_length, (num_true,))
        mask_lengths = torch.randint(
            self.min_mask_length, self.max_mask_length, (num_true,)
        )
        mask_end = mask_start + mask_lengths
        for i, start, end in zip(prob_idx, mask_start, mask_end):
            x[i, start:end, :] = 0
        return x.reshape(*bshape, num_samples, num_channels)


class AddWhiteNoise(nn.Module):
    """Add white noise to the input tensor"""

    def __init__(
        self, min_snr_in_db: float = 0, max_snr_in_db: float = 10, p: float = 0.5
    ):
        super().__init__()
        self.min_snr_in_db = min_snr_in_db
        self.max_snr_in_db = max_snr_in_db
        self.p = p

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Adds white noise to the input tensor

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: Augmented tensor
        """
        # Assuming x has shape (*batch_size, num_samples, num_channels)
        if torch.rand(1) > self.p:
            return x
        snr = (
            torch.rand(*x.shape[:-2]) * (self.max_snr_in_db - self.min_snr_in_db)
            + self.min_snr_in_db
        )  # Shape: (*batch_size)
        snr = snr.to(x.device)

        noise = torch.randn_like(x)
        signal_power = torch.sum(x**2, dim=(-2, -1))  # Shape: (*batch_size)
        orig_noise_power = torch.sqrt(
            torch.sum(noise**2, dim=(-1, -2))
        )  # Shape: (*batch_size)
        new_noise_power = torch.sqrt(
            signal_power / 10 ** (snr / 10)
        )  # Shape: (*batch_size)
        noise = noise * new_noise_power[:, None, None] / orig_noise_power[:, None, None]
        return x + noise


class AddLowFreqNoise(nn.Module):
    def __init__(self, p: float = 0.5):
        super().__init__()
        self.p = p

    def forward(self, audio: torch.Tensor):
        """Adds low frequency noise to an audio tensor of shape (B, T, C)"""
        if torch.rand(1) > self.p:
            return audio

        rand_shape = (audio.shape[0], audio.shape[1] // 2 + 1, audio.shape[2])
        spectrum = torch.rand(rand_shape) * torch.exp(
            -1.0j * torch.rand(rand_shape) * 2 * torch.pi
        )
        spectrum = spectrum.to(audio.device)
        freq_decay = torch.exp(-2 * torch.linspace(0, 1, spectrum.shape[-2])).to(
            audio.device
        )
        freq_decay[0] = 0

        noise = torch.fft.irfft(spectrum * freq_decay[None, :, None], dim=-2)
        noise = (noise - noise.mean()) / noise.std()
        mix = (audio + noise) / sqrt(2)
        return mix


def build_augmentations(CONFIG: dict) -> nn.Module:
    """Builds an augmentation module using the config parameters
    The augmentation module is a subclass of nn.Module which takes in a batched data tensor
    and returns a batched data tensor with the augmentations applied.

    Args:
        CONFIG (dict): Whole config dictionary

    Returns:
        nn.Module: Augmemtation module
    """
    augmentations = []
    aug_config = CONFIG["AUGMENTATIONS"]

    if not CONFIG["DATA"]["AUGMENT_DATA"]:
        return Identity()

    if inversion_config := aug_config.get("INVERSION", False):
        # Inverts the polarity of the audio
        inversion = PolarityInversion(
            p=inversion_config.get("PROB", 0.5),
        )
        augmentations.append(inversion)

    if noise_config := aug_config.get("NOISE", False):
        # Adds white background noise to the audio
        noise = AddWhiteNoise(
            min_snr_in_db=noise_config.get("MIN_SNR", 0),
            max_snr_in_db=noise_config.get("MAX_SNR", 10),
            p=noise_config.get("PROB", 0.5),
        )
        augmentations.append(noise)

    if low_freq_noise_config := aug_config.get("LOW_FREQ_NOISE", False):
        # Adds low frequency noise to the audio
        low_freq_noise = AddLowFreqNoise(
            p=low_freq_noise_config.get("PROB", 0.5),
        )
        augmentations.append(low_freq_noise)

    if mask_config := aug_config.get("MASK", False):
        # Masks a random part of the audio
        mask = TimeMask(
            min_mask_length=mask_config.get("MIN_LENGTH", 75),
            max_mask_length=mask_config.get("MAX_LENGTH", 250),
            prob=mask_config.get("PROB", 0.5),
        )
        augmentations.append(mask)

    return nn.Sequential(*augmentations)



import torch
from torch import nn

from vocalocator.architectures.base import VocalocatorArchitecture
from vocalocator.outputs import ModelOutputFactory


class VocalocatorSimpleLayer(torch.nn.Module):
    def __init__(
        self,
        channels_in: int,
        channels_out: int,
        filter_size: int,
        *,
        downsample: bool,
        dilation: int,
        use_bn: bool = True
    ):
        super(VocalocatorSimpleLayer, self).__init__()
        self.fc = torch.nn.Conv1d(
            channels_in,
            channels_out,
            filter_size,
            stride=(2 if downsample else 1),
            dilation=dilation,
        )
        self.gc = torch.nn.Conv1d(
            channels_in,
            channels_out,
            filter_size,
            stride=(2 if downsample else 1),
            dilation=dilation,
        )
        self.batch_norm = (
            torch.nn.BatchNorm1d(channels_out) if use_bn else nn.Identity()
        )

    def forward(self, x):
        fcx = self.fc(x)
        fcx_activated = torch.tanh(fcx) * 0.95 + fcx * 0.05

        gcx = self.gc(x)
        gcx_activated = torch.sigmoid(gcx)

        prod = fcx_activated * gcx_activated
        return self.batch_norm(prod)


class VocalocatorSimpleNetwork(VocalocatorArchitecture):
    defaults = {
        "USE_BATCH_NORM": True,
        "SHOULD_DOWNSAMPLE": [False, True] * 5,
        "CONV_FILTER_SIZES": [33] * 10,
        "CONV_NUM_CHANNELS": [16, 16, 32, 32, 64, 64, 128, 128, 256, 256],
        "CONV_DILATIONS": [1] * 10,
        "OUTPUT_COV": True,
        "REGULARIZE_COV": False,
    }

    def __init__(self, CONFIG, output_factory: ModelOutputFactory):
        super(VocalocatorSimpleNetwork, self).__init__(CONFIG, output_factory)
        N = CONFIG["DATA"]["NUM_MICROPHONES"]

        # Obtains model-specific parameters from the config file and fills in missing entries with defaults
        model_config = VocalocatorSimpleNetwork.defaults.copy()
        model_config.update(CONFIG.get("MODEL_PARAMS", {}))
        CONFIG["MODEL_PARAMS"] = (
            model_config  # Save the parameters used in this run for backward compatibility
        )

        should_downsample = model_config["SHOULD_DOWNSAMPLE"]
        self.n_channels = model_config["CONV_NUM_CHANNELS"]
        filter_sizes = model_config["CONV_FILTER_SIZES"]
        dilations = model_config["CONV_DILATIONS"]

        min_len = min(
            len(self.n_channels),
            len(filter_sizes),
            len(should_downsample),
            len(dilations),
        )
        self.n_channels = self.n_channels[:min_len]
        filter_sizes = filter_sizes[:min_len]
        should_downsample = should_downsample[:min_len]
        dilations = dilations[:min_len]

        use_batch_norm = model_config["USE_BATCH_NORM"]

        self.n_channels.insert(0, N)

        convolutions = [
            VocalocatorSimpleLayer(
                in_channels,
                out_channels,
                filter_size,
                downsample=downsample,
                dilation=dilation,
                use_bn=use_batch_norm,
            )
            for in_channels, out_channels, filter_size, downsample, dilation in zip(
                self.n_channels[:-1],
                self.n_channels[1:],
                filter_sizes,
                should_downsample,
                dilations,
            )
        ]
        self.conv_layers = torch.nn.Sequential(*convolutions)

        self.final_pooling = nn.AdaptiveAvgPool1d(1)

        if not isinstance(self.n_outputs, int):
            raise ValueError(
                "Number of parameters to output is undefined! Maybe check the model configuration and ModelOutputFactory object?"
            )
        self.coord_readout = torch.nn.Linear(self.n_channels[-1], self.n_outputs)

    def _forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.transpose(
            -1, -2
        )  # (batch, seq_len, channels) -> (batch, channels, seq_len) needed by conv1d
        h1 = self.conv_layers(x)
        h2 = torch.squeeze(self.final_pooling(h1), dim=-1)
        coords = self.coord_readout(h2)
        return coords

    def clip_grads(self):
        nn.utils.clip_grad_norm_(self.parameters(), 1.0, error_if_nonfinite=True)



"""
Common utility functions for things like embedding covariance model output into
an array containing the outputted mean and Cholesky covariance factor.
"""

import torch
from torch.nn import functional as F


def build_cov_output(raw_output, device):
    """
    Given a batch of length-5 vectors, return a batch of (3, 2) arrays
    where for each array, arr[0] is the mean, and arr[1:] is the Cholesky
    factor of the covariance matrix.
    """
    y_hat = raw_output[:, :2]  # (batch, 2)
    # calculate the lower triangular Cholesky factor
    # of the covariance matrix
    len_batch = raw_output.shape[0]
    # initialize an array of zeros into which we'll put
    # the model output
    L = torch.zeros(len_batch, 2, 2, device=device)
    # embed the elements into the matrix
    idxs = torch.tril_indices(2, 2)
    L[:, idxs[0], idxs[1]] = raw_output[:, 2:]
    # apply softplus to the diagonal entries to guarantee the resulting
    # matrix is positive definite
    new_diagonals = F.softplus(L.diagonal(dim1=-2, dim2=-1))
    L = L.diagonal_scatter(new_diagonals, dim1=-2, dim2=-1)
    # reshape y_hat so we can concatenate it to L
    y_hat = y_hat.reshape((-1, 1, 2))  # (batch, 1, 2)
    # concat the two to make a (batch, 3, 2) tensor
    concatenated = torch.cat((y_hat, L), dim=-2)
    return concatenated



"""
Utility functions, like for calibration calculation.
"""

import logging
import math
from typing import Optional, Tuple, Union

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import FigureBase

logger = logging.getLogger(__name__)


def subplots(
    n_plots, scale_factor=4, sharex=True, sharey=True, **kwargs
) -> tuple[FigureBase, list[Axes]]:
    """
    Create nicely sized and laid-out subplots for a desired number of plots.
    """
    # essentially we want to make the subplots as square as possible
    # number of rows is the largest factor of n_plots less than sqrt(n_plots)
    options = range(1, int(math.sqrt(n_plots) + 1))
    n_rows = max(filter(lambda n: n_plots % n == 0, options))
    n_cols = int(n_plots / n_rows)
    # now generate the Figure and Axes pyplot objects
    # cosmetic scale factor to make larger plot
    figsize = (n_cols * scale_factor, n_rows * scale_factor)
    fig, axs = plt.subplots(
        n_rows, n_cols, figsize=figsize, sharex=sharex, sharey=sharey, **kwargs
    )
    flattened_axes = []
    for ax_row in axs:
        if isinstance(ax_row, np.ndarray):
            flattened_axes += list(ax_row)
        else:
            flattened_axes.append(ax_row)
    return fig, flattened_axes


def digitize(locations, bin_edges) -> np.ndarray:
    """
    Wrapper for np.digitize where an error is raised if a value far
    outside the given range is encountered. NOTE: We assume that the
    bins are given in increasing order.
    """
    # check that the bins are in increasing order
    diffs = np.diff(bin_edges)
    if (diffs <= 0).any():
        raise ValueError("Expected array `bins` to be in increasing order.")
    # get max distance between bins
    max_dx = diffs.max()
    # define a new bin array where the highest bin
    # is bins[-1] + tol
    # say values less than this are in the highest bin.
    # this is to catch floating point errors that push values
    # greater than bin_edges[-1], while still
    # letting us catch extreme values that are way too high.
    tol = max_dx
    extended_bins = np.append(bin_edges, bin_edges[-1] + tol)
    # digitize locations using these new bins, removing the
    # leftmost bin edge to avoid off-by-one errors.
    edges_to_use = extended_bins[1:]
    bin_idxs = np.digitize(locations, edges_to_use)
    # if any value was greater than bins[-1] + max_dx,
    # raise a value error.
    if (bin_idxs == len(edges_to_use)).any():
        pass
        # positions = bin_idxs == len(edges_to_use)
        # values = locations[positions]
        # err_display = [
        #     f"idx: {p} | value: {v}" for (p, v) in zip(positions.nonzero(), values)
        # ]
        # raise ValueError(
        #     f"Encountered value far greater than the largest bin edge! "
        #     f"Largest bin edge: {bin_edges[-1]}; Invalid values and their "
        #     f"positions: {err_display}"
        # )
    # if not, say that the values were sufficiently close to the bin edges
    # and clip them to match the number of bins
    num_bins = len(bin_edges) - 1
    highest_bin_idx = num_bins - 1
    return bin_idxs.clip(0, highest_bin_idx)


def assign_to_bin_2d(locations, xgrid, ygrid):
    """
    Return an array of indices of the 2d bins to which each input in
    `locations` corresponds.

    The indices correspond to the "flattened" version of the grid. In essence,
    for a point in bin (i, j), the output is i + (n_x_bins * j), where n_x_bins
    is the number of grid bins in the x direction--essentially the number
    of gridpoints in that direction - 1.
    """
    # locations: (NUM_SAMPLES, 2)
    # xgrid: (n_y_pts, n_x_pts)
    # xgrid: (n_y_pts, n_x_pts)
    x_coords = locations[:, 0]
    y_coords = locations[:, 1]
    # 1d array of numbers representing x coord of each bin
    x_bin_edges = xgrid[0]
    # same for y coord
    y_bin_edges = ygrid[:, 0]
    # assign each coord to a bin in one dimension
    x_idxs = digitize(x_coords, x_bin_edges)
    y_idxs = digitize(y_coords, y_bin_edges)
    # NOTE: we expect model output to have shape (NUM_SAMPLES, n_y_bins, n_x_bins)
    # where n_y_bins = len(y_bin_edges) - 1, and similar for n_x_bins.
    # so when we flatten, the entry at coordinate (i, j) gets mapped to
    # (n_x_bins * j) + i
    n_x_bins = len(x_bin_edges) - 1
    return (n_x_bins * y_idxs) + x_idxs


def make_xy_grids(
    arena_dims: Union[Tuple[float, float], np.ndarray],
    resolution: Optional[float] = None,
    shape: Optional[Union[Tuple[float, float], np.ndarray]] = None,
    return_center_pts: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given a two-tuple or np array storing the x and y dimensions and a desired
    grid resolution, return a tuple of arrays stooring the x and y coordinates
    at every point in the arena spaced apart by `desired_resolution`.

    Optionally, calculate those gridpoints and instead return the CENTER of each
    bin based on the flag `return_center_pts`.

    Note that this function expects shape in the format (n_y_pts, n_x_pts).

    Examples:
    ```
    >>> make_xy_grids((4, 3), shape=(2, 3))
    (
        array([
            [0., 2., 4.],
            [0., 2., 4.]
        ]),
        array([
            [0., 0., 0.],
            [3., 3., 3.]
        ])
    )
    >>> make_xy_grids((4, 3), shape=(1, 2), return_center_pts=True)
    (
        array([
            [0.66666667, 2.        , 3.33333333],
            [0.66666667, 2.        , 3.33333333]
        ]),
        array([
            [0.75, 0.75, 0.75],
            [2.25, 2.25, 2.25]
        ])
    )
    ```

    """
    if resolution is None and shape is None:
        raise ValueError("One of `resolution`, `shape` is required!")

    if not resolution:
        # np.meshgrid returns a shape of (n_y_pts, n_x_pts)
        # but expects (xs, ys) as arguments.
        # reverse the shape so we match this convention.
        pts_per_dim = np.array(shape)[::-1]
    else:
        pts_per_dim = np.ceil(np.array(arena_dims) / resolution).astype(int)

    def _coord_array(dim_pts):
        """
        Get an array of coordinates along one axis.

        Expects `dim_pts` to be a tuple (dim, n_pts), where `dim`
        is the length of the grid along the current axis, and `n_pts`
        is the desired number of points to be placed along the axis.
        """
        dimension, n_pts = dim_pts
        # if the user requested to return the CENTER of each bin,
        # create one extra gridpoint in each direction
        # then return the average of each successive bin
        if return_center_pts:
            edge_coords = np.linspace(-dimension / 2, dimension / 2, n_pts + 1)
            # add half the successive differences to get avgs
            # between edge_pts[i] and edge_pts[i+1]
            coords = edge_coords[:-1] + (np.diff(edge_coords) / 2)
        else:
            coords = np.linspace(-dimension / 2, dimension / 2, n_pts)
        return coords

    xs, ys = map(_coord_array, zip(arena_dims, pts_per_dim))

    xgrid, ygrid = np.meshgrid(xs, ys)

    return (xgrid, ygrid)



"""
Functions to construct Datasets and DataLoaders for training and inference
"""

import os
from pathlib import Path
from typing import Optional, Tuple, Union

import h5py
import numpy as np
import torch
from scipy.io import wavfile
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset
from torchaudio import functional as AF


class VocalizationDataset(Dataset):
    def __init__(
        self,
        datapath: Union[Path, str],
        crop_length: int,
        *,
        nodes: Optional[list[str]] = None,
        inference: bool = False,
        arena_dims: Optional[Union[np.ndarray, Tuple[float, float]]] = None,
        index: Optional[np.ndarray] = None,
        normalize_data: bool = True,
        sample_rate: int = 192000,
        sample_vocalization_dir: Optional[Path] = None,
    ):
        """
        Args:
            datapath (Path): Path to HDF5 dataset
            inference (bool, optional): When true, data will be cropped deterministically. Defaults to False.
            crop_length (int): Length of audio samples to return.
            arena_dims (Optional[Union[np.ndarray, Tuple[float, float]]], optional): Dimensions of the arena in mm. Used to scale labels.
            index (Optional[np.ndarray], optional): An array of indices to use for this dataset. Defaults to None, which will use the full dataset
        """
        if isinstance(datapath, str):
            datapath = Path(datapath)
        self.datapath = datapath
        dataset = h5py.File(self.datapath, "r")

        # dataset cannot exist as a member of the object until after pytorch has cloned and
        # spread the dataset objects across multiple processes.
        # This is because h5py handles cannot be pickled and pytorch uses pickle under the hood
        # I get around this by re-initializing the h5py.File lazily in __getitem__
        self.dataset: Optional[h5py.File] = None

        if not isinstance(arena_dims, np.ndarray):
            arena_dims = np.array(arena_dims).astype(np.float32)

        if "length_idx" not in dataset and "rir_length_idx" not in dataset:
            raise ValueError("Improperly formatted dataset")

        if "audio" in dataset:
            self.is_rir_dataset = False
            self.length = len(dataset["length_idx"]) - 1
        elif "rir" in dataset:
            self.is_rir_dataset = True
            self.length = len(dataset["rir_length_idx"]) - 1
        else:
            raise ValueError("Improperly formatted dataset")

        self.inference = inference
        self.arena_dims = arena_dims
        self.crop_length = crop_length
        self.index = index
        self.normalize_data = normalize_data
        self.sample_rate = sample_rate
        self.sample_vocalization_dir = sample_vocalization_dir

        # Determine which nodes indices to select
        if nodes is None:
            self.node_indices = np.array([0])
        else:
            if "node_names" not in dataset:
                pass
                # raise ValueError("Dataset does not contain node names")
                dset_node_names = []
            else:
                dset_node_names = list(map(bytes.decode, dataset["node_names"]))
            self.node_indices = [
                i for i, node in enumerate(dset_node_names) if node in nodes
            ]

            if not self.node_indices:
                self.node_indices = np.array([0])
                # raise ValueError(
                # "No nodes found in dataset with the given names: {}".format(nodes)
                # )
            self.node_indices = np.array(self.node_indices)

        if self.index is not None:
            self.length = len(self.index)

        if self.is_rir_dataset and self.sample_vocalization_dir is None:
            raise ValueError("RIR dataset requires sample vocalizations")

        worker_info = torch.utils.data.get_worker_info()
        seed = 0 if worker_info is None else worker_info.seed
        self.rng = np.random.default_rng(seed)

        self.sample_vocalizations = []
        if self.is_rir_dataset:
            for wavfile_path in Path(self.sample_vocalization_dir).glob("*.wav"):
                fs, data = wavfile.read(wavfile_path)
                if fs != sample_rate:
                    continue
                data = self.convert_audio_to_float(data)
                if data is None:
                    continue
                if len(data.shape) > 1:
                    # scipy reads stereo files as (n_samples, n_channels)
                    channel_powers = np.sum(data**2, axis=0)
                    data = data[:, np.argmax(channel_powers)]
                # Don't want some vocalizations to be louder than others
                data = (data - data.mean()) / data.std()
                self.sample_vocalizations.append(torch.from_numpy(data).float())

            if not self.sample_vocalizations:
                raise ValueError("No valid vocalizations found")
        dataset.close()

    def convert_audio_to_float(self, audio: np.ndarray):
        """Converts audio to float32 and scales it to the range [-1, 1]"""
        float_types = (np.float32, np.float64)
        if audio.dtype in float_types:
            return audio.astype(np.float32)

        int_types = (np.int8, np.int16, np.int32)
        if audio.dtype not in int_types:
            return None
        iinfo = np.iinfo(audio.dtype)
        imin, imax = iinfo.min, iinfo.max
        return (audio.astype(np.float64) / (imax - imin)).astype(np.float32)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        if self.dataset is None:
            self.dataset = h5py.File(self.datapath, "r")
        true_idx = idx
        if self.index is not None:
            true_idx = self.index[idx]
        return self.__processed_data_for_index__(true_idx)

    def __make_crop(self, audio: torch.Tensor, crop_length: int):
        """Given an audio sample of shape (n_samples, n_channels), return a random crop
        of shape (crop_length, n_channels)
        """
        audio_len, _ = audio.shape
        valid_range = audio_len - crop_length
        if valid_range <= 0:  # Audio is shorter than desired crop length, pad right
            pad_size = crop_length - audio_len
            # will fail if input is numpy array
            return F.pad(audio, (0, 0, 0, pad_size))
        if self.inference:
            range_start = 0  # Ensure the output is deterministic at inference time
        else:
            range_start = self.rng.integers(0, valid_range)
        range_end = range_start + crop_length
        return audio[range_start:range_end, :]

    def __rir_for_index(self, idx: int):
        if self.is_rir_dataset:
            start, end = self.dataset["rir_length_idx"][idx : idx + 2]
            return torch.from_numpy(self.dataset["rir"][start:end, ...]).float()
        raise ValueError("Dataset does not contain RIRs")

    def __audio_for_index(self, idx: int):
        """Gets an audio sample from the dataset. Will determine the format
        of the dataset and handle it appropriately.
        """
        start, end = self.dataset["length_idx"][idx : idx + 2]
        audio = self.dataset["audio"][start:end, ...]
        return torch.from_numpy(audio.astype(np.float32))

    def __label_for_index(self, idx: int):
        if "locations" not in self.dataset:
            return None
        if len(self.dataset["locations"].shape) == 2:
            locs = torch.from_numpy(
                self.dataset["locations"][idx].astype(np.float32)[None, :]
            )
            return locs

        locs = self.dataset["locations"][idx, ..., self.node_indices, :]
        locs = torch.from_numpy(locs.astype(np.float32))
        return locs

    def scale_features(
        self,
        audio: np.ndarray,
        labels: np.ndarray,
    ):
        """Scales the inputs to have zero mean and unit variance. Labels are scaled
        from millimeter units to an arbitrary unit with range [-1, 1].
        """

        if labels is not None:
            # Shift range to [-1, 1], but keep units same across dimensions
            # Origin remains at center of arena
            scale_factor = self.arena_dims.max() / 2
            labels /= scale_factor

        if self.normalize_data:
            scaled_audio = (audio - audio.mean()) / audio.std()

        return scaled_audio, labels

    def __processed_data_for_index__(self, idx: int):
        if self.is_rir_dataset:
            # rir shape: (n_channels, n_samples)
            rir = self.__rir_for_index(idx)
            # sample_vocalization shape: (n_samples,)
            sample_vocalization = self.sample_vocalizations[
                self.rng.integers(len(self.sample_vocalizations))
            ]

            sound = AF.convolve(rir.T, sample_vocalization[None, :], mode="full").T
        else:
            sound = self.__audio_for_index(idx)

        sound = self.__make_crop(sound, self.crop_length)
        location = self.__label_for_index(idx)

        sound, location = self.scale_features(sound, location)

        return sound, location

    def collate(self, batch) -> Tuple[torch.Tensor, torch.Tensor]:
        """Collate function for the dataloader. Takes a list of (audio, label) tuples and returns
        a batch of audio and labels.
        """
        audio, labels = [x[0] for x in batch], [x[1] for x in batch]
        audio = torch.stack(audio)
        if labels[0] is not None:
            labels = torch.stack(labels)
        else:
            labels = [None] * len(audio)
        return audio, labels


def build_dataloaders(
    path_to_data: Union[Path, str], config: dict, index_dir: Optional[Path]
) -> tuple[DataLoader, DataLoader, Optional[DataLoader]]:
    # Construct Dataset objects.
    if not isinstance(path_to_data, Path):
        path_to_data = Path(path_to_data)

    arena_dims = config["DATA"]["ARENA_DIMS"]
    batch_size = config["DATA"]["BATCH_SIZE"]
    crop_length = config["DATA"]["CROP_LENGTH"]
    normalize_data = config["DATA"].get("NORMALIZE_DATA", True)
    sample_rate = config["DATA"].get("SAMPLE_RATE", 192000)
    node_names = config["DATA"].get("NODES_TO_LOAD", None)

    vocalization_dir = config["DATA"].get(
        "VOCALIZATION_DIR",
        None,
    )

    index_arrays = {"train": None, "val": None, "test": None}
    if index_dir is not None:
        index_arrays["train"] = np.load(index_dir / "train_set.npy")
        index_arrays["val"] = np.load(index_dir / "val_set.npy")
        if (index_dir / "test_set.npy").exists():
            index_arrays["test"] = np.load(index_dir / "test_set.npy")

    try:
        # I think this function only exists on linux
        avail_cpus = max(1, len(os.sched_getaffinity(0)) - 1)
    except:
        avail_cpus = max(1, os.cpu_count() - 1)

    if path_to_data.is_dir():
        train_path = path_to_data / "train_set.h5"
        val_path = path_to_data / "val_set.h5"
        test_path = path_to_data / "test_set.h5"
    else:
        train_path = path_to_data
        val_path = path_to_data
        test_path = path_to_data
        if index_dir is None:
            # manually create train/val split
            with h5py.File(train_path, "r") as f:
                if "length_idx" in f:
                    dset_size = len(f["length_idx"]) - 1
                elif "rir_length_idx" in f:
                    dset_size = len(f["rir_length_idx"]) - 1
                else:
                    raise ValueError("Improperly formatted dataset")
            full_index = np.arange(dset_size)
            rng = np.random.default_rng(0)
            rng.shuffle(full_index)
            index_arrays["train"] = full_index[: int(0.8 * dset_size)]
            index_arrays["val"] = full_index[
                int(0.8 * dset_size) : int(0.9 * dset_size)
            ]
            index_arrays["test"] = full_index[int(0.9 * dset_size) :]

    traindata = VocalizationDataset(
        train_path,
        arena_dims=arena_dims,
        crop_length=crop_length,
        index=index_arrays["train"],
        normalize_data=normalize_data,
        sample_rate=sample_rate,
        sample_vocalization_dir=vocalization_dir,
        nodes=node_names,
    )

    valdata = VocalizationDataset(
        val_path,
        arena_dims=arena_dims,
        crop_length=crop_length,
        inference=True,
        index=index_arrays["val"],
        normalize_data=normalize_data,
        sample_rate=sample_rate,
        sample_vocalization_dir=vocalization_dir,
        nodes=node_names,
    )

    train_dataloader = DataLoader(
        traindata,
        batch_size=batch_size,
        shuffle=True,
        num_workers=avail_cpus,
        collate_fn=traindata.collate,
    )

    val_dataloader = DataLoader(
        valdata,
        batch_size=batch_size,
        num_workers=avail_cpus,
        shuffle=False,
        collate_fn=valdata.collate,
    )

    test_dataloader = None
    if test_path.exists():
        testdata = VocalizationDataset(
            test_path,
            arena_dims=arena_dims,
            crop_length=crop_length,
            inference=True,
            index=index_arrays["test"],
            normalize_data=normalize_data,
            sample_rate=sample_rate,
            sample_vocalization_dir=vocalization_dir,
        )
        test_dataloader = DataLoader(
            testdata,
            batch_size=batch_size,
            num_workers=1,
            shuffle=False,
            collate_fn=testdata.collate,
        )

    return train_dataloader, val_dataloader, test_dataloader



import json
import logging
import os
from pathlib import Path
from typing import NewType, Optional

import numpy as np
import torch
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    ExponentialLR,
    ReduceLROnPlateau,
    SequentialLR,
)

from ..calibration import CalibrationAccumulator
from ..outputs.base import ModelOutput, ProbabilisticOutput, Unit
from ..training.augmentations import build_augmentations
from ..training.dataloaders import build_dataloaders
from ..training.logger import ProgressLogger
from ..training.models import build_model

JSON = NewType("JSON", dict)


def make_logger(filepath: str) -> logging.Logger:
    logger = logging.getLogger("train_log")
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.FileHandler(filepath))
    logger.addHandler(logging.StreamHandler())
    return logger


class Trainer:
    """A helper class for training and performing inference with Vocalocator models"""

    @staticmethod
    def __query_mem_usage():
        if not torch.cuda.is_available():
            return ""
        used_gb = torch.cuda.max_memory_allocated() / (2**30)
        total_gb = torch.cuda.get_device_properties(0).total_memory / (2**30)
        torch.cuda.reset_peak_memory_stats()
        return "Max mem. usage: {:.2f}/{:.2f}GiB".format(used_gb, total_gb)

    def __init__(
        self,
        data_dir: str,
        model_dir: str,
        config_data: JSON,
        *,
        eval_mode: bool = False,
        index_dir: Optional[Path] = None,
    ):
        """Parameters:
        - data_dir:
            Path to directory containing train/test/val files named train_set.h5, etc...
        - model_dir:
            Path to the directory that will hold model weights and logs
        - config_data:
            Contents of model config as a JSON object (python dictionary-like)
        - eval_mode:
            Should be false when training and true when performing inference.
        """
        self.__eval = eval_mode
        self.__datafile = data_dir
        self.__model_dir = model_dir
        self.__config = config_data

        if self.__config["GENERAL"]["DEVICE"] == "CPU":
            self.device = torch.device("cpu")
        elif torch.cuda.is_available():
            self.device = torch.device("cuda")
        elif torch.backends.mps.is_available():
            self.device = torch.device("mps")
        else:
            # fall back
            self.device = torch.device("cpu")

        print(f"Using device: {self.device}")

        if not self.__eval:
            self.__init_dataloaders(index_dir)
            self.__init_output_dir()
        self.__init_model()

        self.augment = build_augmentations(self.__config)

        if not self.__eval:
            self.__logger.info(f" ==== STARTING TRAINING ====\n")
            self.__logger.info(
                f">> SAVING INITIAL MODEL WEIGHTS TO {self.__init_weights_file}"
            )
            self.save_weights(self.__init_weights_file)

            self.__best_loss = float("inf")

    def save_weights(self, weight_path: str):
        torch.save(self.model.state_dict(), weight_path)

    def finalize(self):
        self.__logger.info(
            ">> FINISHED ALL EPOCHS. TOTAL TIME ELAPSED: "
            + self.__progress_log.print_time_since_initialization()
        )
        self.__logger.info(
            f">> SAVING FINAL MODEL WEIGHTS TO {self.__final_weights_file}"
        )
        self.save_weights(self.__final_weights_file)

    def __init_logger(self):
        log_filepath = os.path.join(self.__model_dir, "train_log.txt")
        self.__logger = make_logger(log_filepath)
        print(f"Saving logs to file: `{log_filepath}`")

    def __init_output_dir(self):
        self.__best_weights_file = os.path.join(self.__model_dir, "best_weights.pt")
        self.__init_weights_file = os.path.join(self.__model_dir, "init_weights.pt")
        self.__final_weights_file = os.path.join(self.__model_dir, "final_weights.pt")

        # Found that it's helpful to keep track of this
        self.__config["DATA"]["DATAFILE_PATH"] = self.__datafile

        self.__init_logger()

        if self.__traindata is not None:
            self.__logger.info(f"Training set:\t{self.__traindata.dataset}")
        if self.__valdata is not None:
            self.__logger.info(f"Validation set:\t{self.__valdata.dataset}")

        self.num_epochs: int = self.__config["OPTIMIZATION"]["NUM_EPOCHS"]
        self.__progress_log = ProgressLogger(
            self.num_epochs,
            self.__traindata,
            self.__valdata,
            self.__config["GENERAL"]["LOG_INTERVAL"],
            self.__model_dir,
            self.__logger,
        )

    def __init_dataloaders(self, index_dir: Optional[Path]):
        # Load training set, validation set, test set.
        self.__traindata, self.__valdata, testdata = build_dataloaders(
            self.__datafile, self.__config, index_dir
        )

        # save indices for reproducibility
        train_idx = self.__traindata.dataset.index
        val_idx = self.__valdata.dataset.index

        if train_idx is not None:
            index_dir = Path(self.__model_dir) / "indices"
            index_dir.mkdir(parents=True, exist_ok=True)
            np.save(index_dir / "train_set.npy", train_idx)
            np.save(index_dir / "val_set.npy", val_idx)

        if testdata is not None:
            test_idx = testdata.dataset.index
            np.save(index_dir / "test_set.npy", test_idx)

    def __init_model(self):
        """Creates the model, optimizer, and loss function."""
        # Set random seeds.
        torch.manual_seed(self.__config["GENERAL"]["TORCH_SEED"])
        np.random.seed(self.__config["GENERAL"]["NUMPY_SEED"])

        # Specify network architecture and loss function.
        self.model, self.__loss_fn = build_model(self.__config)
        if not self.__eval:
            self.__config["WEIGHTS_PATH"] = self.__best_weights_file
        self.model.to(self.device)

        # In inference mode, there is no logger
        if not self.__eval:
            with open(os.path.join(self.__model_dir, "config.json"), "w") as ctx:
                json.dump(self.__config, ctx, indent=4)

            self.__logger.info(self.model.__repr__())

            optim_config = self.__config["OPTIMIZATION"]
            if optim_config["OPTIMIZER"] == "SGD":
                base_optim = torch.optim.SGD
                optim_args = {
                    "momentum": optim_config["MOMENTUM"],
                    "weight_decay": optim_config["WEIGHT_DECAY"],
                }
            elif optim_config["OPTIMIZER"] == "ADAM":
                base_optim = torch.optim.Adam
                optim_args = {"betas": optim_config["ADAM_BETAS"]}
            else:
                raise NotImplementedError(
                    f'Unrecognized optimizer "{optim_config["OPTIMIZER"]}"'
                )

            params = (
                self.model.trainable_params()
                if hasattr(self.model, "trainable_params")
                else self.model.parameters()
            )

            self.__optim = base_optim(
                params,
                lr=optim_config["INITIAL_LEARNING_RATE"],
                **optim_args,
            )

            scheduler_configs = optim_config["SCHEDULERS"]

            schedulers = []
            epochs_active_per_scheduler = []

            for scheduler_config in scheduler_configs:
                scheduler_type = scheduler_config["SCHEDULER_TYPE"]
                # by default, if number of active epochs is not specified, default to
                # running for the remaining duration.
                epochs_active = scheduler_config.get("NUM_EPOCHS_ACTIVE")
                if epochs_active is None:
                    total_specified_already = sum(epochs_active_per_scheduler)
                    remaining_dur = self.num_epochs - total_specified_already
                    self.__logger.info(
                        "No `NUM_EPOCHS_ACTIVE` parameter passed to scheduler "
                        f"{scheduler_type}! Defaulting to remaining train duration, "
                        f"{remaining_dur}."
                    )
                    epochs_active = remaining_dur
                epochs_active_per_scheduler.append(epochs_active)

                # parse lr scheduler
                if scheduler_type == "COSINE_ANNEALING":
                    base_scheduler = CosineAnnealingLR
                    scheduler_args = {
                        "T_max": epochs_active,
                        "eta_min": scheduler_config.get("MIN_LEARNING_RATE", 0),
                    }
                elif scheduler_type == "EXPONENTIAL_DECAY":
                    base_scheduler = ExponentialLR
                    scheduler_args = {
                        "gamma": scheduler_config["MULTIPLICATIVE_DECAY_FACTOR"]
                    }
                elif scheduler_type == "REDUCE_ON_PLATEAU":
                    base_scheduler = ReduceLROnPlateau
                    scheduler_args = {
                        "factor": scheduler_config.get(
                            "MULTIPLICATIVE_DECAY_FACTOR", 0.1
                        ),
                        "patience": scheduler_config.get("PLATEAU_DECAY_PATIENCE", 10),
                        "threshold_mode": scheduler_config.get(
                            "PLATEAU_THRESHOLD_MODE", "rel"
                        ),
                        "threshold": scheduler_config.get("PLATEAU_THRESHOLD", 1e-4),
                        "min_lr": scheduler_config.get("MIN_LEARNING_RATE", 0),
                    }
                else:
                    raise NotImplementedError(
                        f'Unrecognized scheduler "{scheduler_config["SCHEDULER_TYPE"]}"'
                    )
                schedulers.append(base_scheduler(self.__optim, **scheduler_args))

            # sequential lr expects the points at which it should switch,
            # so take the cumulative sum and throw out the endpoint
            milestones = list(np.cumsum(epochs_active_per_scheduler))[:-1]
            self.__scheduler = SequentialLR(
                self.__optim, schedulers=schedulers, milestones=milestones
            )

    def train_epoch(self):
        # Set the learning rate using cosine annealing.
        self.__progress_log.start_training()
        self.model.train()

        iter = 0
        for sounds, locations in self.__traindata:
            # Move data to device
            sounds = sounds.to(self.device)
            locations = locations.to(self.device)

            # This should always exist, but might be the identity function
            sounds = self.augment(sounds)

            # Prepare optimizer.
            self.__optim.zero_grad()

            # Forward pass
            outputs = self.model(sounds)

            # Compute loss.
            losses = self.__loss_fn(outputs, locations)
            mean_loss = torch.mean(losses)

            # Backwards pass.
            mean_loss.backward()
            if self.__config["OPTIMIZATION"]["CLIP_GRADIENTS"]:
                self.model.clip_grads()
            self.__optim.step()

            # Count batch as completed.
            self.__progress_log.log_train_batch(
                mean_loss.item(), np.nan, sounds.shape[0]
            )
            iter += 1
            # if iter > 10: break

    def eval_validation(self):
        self.__progress_log.start_testing()
        self.model.eval()
        if self.__valdata is not None:
            with torch.no_grad():
                idx = 0
                compute_calibration = False
                for sounds, locations in self.__valdata:
                    sounds = sounds.to(self.device)

                    batch_err = 0

                    outputs: list[ModelOutput] = self.model(sounds, unbatched=True)

                    for output, location in zip(outputs, locations):
                        # Calculate error in cm
                        point_estimate = (
                            output.point_estimate(units=Unit.CM).cpu().numpy()
                        )  # Returns (n_node, n_dim)
                        location = (
                            output._convert(location, Unit.ARBITRARY, Unit.CM)
                            .float()
                            .cpu()
                            .numpy()
                        )

                        # Ensure the same number of dimensions are being used
                        n_nodes: int = output.nnode
                        n_dims: int = output.ndim
                        location = location[..., :n_nodes, :n_dims]

                        err = np.linalg.norm(
                            point_estimate - location, axis=-1
                        )  # (n_node,)
                        err = err[0]  # Only look at the first node
                        batch_err += err

                        if output.computes_calibration:
                            compute_calibration = True
                            if idx == 0:
                                ca = CalibrationAccumulator(
                                    output.arena_dims[Unit.MM].cpu().numpy()
                                )
                            location_mm = location * 10
                            ca.calculate_step(output, location_mm)

                        idx += 1

                    # Log progress
                    self.__progress_log.log_val_batch(
                        batch_err / sounds.shape[0],
                        np.nan,
                        sounds.shape[0],
                    )

            # Done with epoch.
            cal_curve = None
            if compute_calibration:
                cal_curve = ca.results()["calibration_curve"]
            val_loss = self.__progress_log.finish_epoch(
                calibration_curve=cal_curve
            )  # will have shape (n_node,)
            reported_val_loss = ",".join([f"{float(x):.3f}" for x in val_loss])
            val_loss = val_loss.mean()

        else:
            val_loss = 0.0

        # Save best set of weights.
        if val_loss < self.__best_loss or self.__valdata is None:
            self.__best_loss = val_loss
            fmt_val_loss = f"{reported_val_loss}cm"
            self.__logger.info(
                f">> MEAN VALIDATION LOSS, {fmt_val_loss}, IS BEST SO FAR, SAVING WEIGHTS TO {self.__best_weights_file}"
            )
            self.save_weights(self.__best_weights_file)
        self.__logger.info(Trainer.__query_mem_usage())

        return val_loss

    def update_learning_rate(self, val_loss: float):
        """
        Update the learning rate after epoch completion.
        """
        if isinstance(self.__scheduler, ReduceLROnPlateau):
            self.__scheduler.step(val_loss)
        else:
            self.__scheduler.step()

    def train(self):
        """
        Train the model for self.num_epochs passes through the dataset.
        """
        for _ in range(self.num_epochs):
            self.train_epoch()
            val_loss = self.eval_validation()
            # Update learning rates based on the mean val loss across nodes
            self.update_learning_rate(val_loss)
        self.finalize()



import copy

SIMPLENET_BASE = {
    "MOMENTUM": 0.9,
    "TRAIN_BATCH_SIZE": 64,
    "VAL_BATCH_SIZE": 64,
    "WEIGHT_DECAY": 1e-05,
    "USE_BATCH_NORM": True,
    "MAX_LEARNING_RATE": 1e-03,
    "MIN_LEARNING_RATE": 1e-05,
    "SAMPLE_LEN": 8192,
    "SHOULD_DOWNSAMPLE": [False, True, True, True, True, True, False],
    "CONV_FILTER_SIZES": [19, 7, 39, 41, 23, 29, 33],
    "NUM_MICROPHONES": 4,
    "NUM_EPOCHS": 20,
    "ARCHITECTURE": "VocalocatorSimpleNetwork",
    "COMPUTE_XCORRS": True,
    "AUGMENT_DATA": False,
    "CLIP_GRADIENTS": True,
    "CONV_NUM_CHANNELS": [16, 16, 16, 32, 32, 32, 64, 64, 64],
    "CONV_DILATIONS": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
    "ARENA_WIDTH": 600,
    "ARENA_LENGTH": 400,
    "CONFIG_NAME": "simplenet_no_cov",
    "DEVICE": "CPU",
    "TORCH_SEED": 888,
    "NUMPY_SEED": 777,
    "LOG_INTERVAL": 3,
    "NUM_AUDIO_SAMPLES": 10000,
    "AUDIO_SAMPLE_RATE": 125000,
    "SAVE_SAMPLE_OUTPUT": False,
    "SAVE_LOSS_PLOT": True,
    "OPTIMIZER": "SGD",
    "TEST_BATCH_SIZE": 64,
    "POOLING": "AVG",
    "AUGMENT_STRETCH_MIN": 0.95,
    "AUGMENT_STRETCH_MAX": 1.1,
    "AUGMENT_STRETCH_PROB": 1e-06,
    "AUGMENT_GAUSS_NOISE": 0.005,
    "AUGMENT_PITCH_MIN": -1.0,
    "AUGMENT_PITCH_MAX": 1.0,
    "AUGMENT_PITCH_PROB": 1e-06,
    "AUGMENT_SHIFT_MIN": -0.1,
    "AUGMENT_SHIFT_MAX": 0.1,
    "AUGMENT_SHIFT_PROB": 1.0,
    "AUGMENT_INVERSION_PROB": 0.5,
    "AUGMENT_LABELS": False,
    "AUGMENT_FLIP_HORIZ": True,
    "AUGMENT_FLIP_VERT": True,
    "DATAFILE_PATH": "/mnt/home/atanelus/ceph/iteration/small_room_4/",
    "WEIGHTS_PATH": "/mnt/home/achoudhri/ceph/end_to_end/speaker/attempt_1_2142023/2296261/pretrain/best_weights.pt",
}

SIMPLENET_COV = copy.deepcopy(SIMPLENET_BASE)
SIMPLENET_COV["OUTPUT_COV"] = True

ENSEMBLE = {
    "MODELS": [SIMPLENET_COV, SIMPLENET_COV, SIMPLENET_COV],
    "ARCHITECTURE": "VocalocatorEnsemble",
    "ARENA_WIDTH": 600,
    "ARENA_LENGTH": 400,
    "CONFIG_NAME": "simplenet_no_cov",
    "DEVICE": "CPU",
}

ENSEMBLE_MISSING_COV = {
    "MODELS": [SIMPLENET_COV, SIMPLENET_BASE, SIMPLENET_COV],
    "ARCHITECTURE": "VocalocatorEnsemble",
    "ARENA_WIDTH": 600,
    "ARENA_LENGTH": 400,
    "CONFIG_NAME": "simplenet_no_cov",
    "DEVICE": "CPU",
}



"""Initialize model and loss function from configuration."""

import json
from pathlib import Path
from typing import Any, Callable, Union

import numpy as np
import torch

from vocalocator.outputs import (
    GaussianOutputFixedVariance,
    ModelOutput,
    ModelOutputFactory,
    ProbabilisticOutput,
)
from vocalocator.outputs.base import BaseDistributionOutput, MDNOutput

from ..architectures.base import VocalocatorArchitecture
from ..architectures.ensemble import VocalocatorEnsemble
from .losses import negative_log_likelihood, squared_error


def __subclasses_recursive(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in __subclasses_recursive(c)]
    )


def subclasses(cls):
    return {
        c.config_name: c
        for c in list(__subclasses_recursive(cls))
        if hasattr(c, "config_name")
    }


ARCHITECTURES = {
    model.__name__.lower(): model for model in VocalocatorArchitecture.__subclasses__()
}

OUTPUT_TYPES = subclasses(ModelOutput)


def make_output_factory(config: dict[str, Any]) -> ModelOutputFactory:
    """
    Parse the provided config and return an appropriately configured
    ModelOutputFactory instance.
    """
    model_params = config["MODEL_PARAMS"]
    provided_type = model_params.get("OUTPUT_TYPE", "").upper()
    if provided_type not in OUTPUT_TYPES:
        raise ValueError(
            f"Unrecognized output type: {provided_type}! "
            f"Allowable types are: {list(OUTPUT_TYPES.keys())}"
        )
    output_type = OUTPUT_TYPES[provided_type]

    arena_dims = torch.tensor(config["DATA"]["ARENA_DIMS"])
    arena_dims_units = config["DATA"].get("ARENA_DIMS_UNITS")

    if not arena_dims_units:
        raise ValueError(
            "Need to specify units for arena dimensions under key "
            "`ARENA_DIMS_UNITS`!"
        )

    # now handle certain subclasses which require additional information
    # to be parsed from config
    additional_kwargs = {}
    if output_type == MDNOutput:
        # check for additional parameter 'CONSTITUENT_DISTRIBUTIONS'
        if not model_params.get("CONSTITUENT_DISTRIBUTIONS"):
            raise ValueError(
                "Given mixture density network output type (MDNOutput), "
                "expected parameter `CONSTITUENT_DISTRIBUTIONS` expressing "
                "the component distributions for the MDN. This argument "
                "should be in the `MODEL_PARAMS` subdict of the config "
                "and should contain a list of valid output type strings."
            )
        provided_constituents = model_params["CONSTITUENT_DISTRIBUTIONS"]

        # only allow subclasses of BaseDistributionOutput now
        base_distr_subclasses = subclasses(BaseDistributionOutput)
        constituent_types = []
        for constituent_type in provided_constituents:
            if constituent_type not in base_distr_subclasses:
                raise ValueError(
                    f"Unrecognized output type as constituent for MDN: {constituent_type}! "
                    f"Allowable types are: {list(base_distr_subclasses.keys())}"
                )
            constituent_types.append(base_distr_subclasses[constituent_type])

        additional_kwargs["constituent_response_types"] = constituent_types
        # for more details on this config param, see the MDNOutput constructor
        additional_kwargs["constituent_extra_kwargs"] = model_params.get(
            "CONSTITUENT_EXTRA_KWARGS", {}
        )

    elif output_type == GaussianOutputFixedVariance:
        # expect 'VARIANCE' and 'VARIANCE_UNITS'
        variance = model_params.get("VARIANCE", "")
        units = model_params.get("VARIANCE_UNITS", "")

        additional_kwargs["variance"] = torch.tensor(variance)
        additional_kwargs["units"] = units

    return ModelOutputFactory(
        arena_dims=arena_dims,
        arena_dim_units=arena_dims_units,
        output_type=output_type,
        **additional_kwargs,
    )


LossFunction = Callable[[ModelOutput, torch.Tensor], torch.Tensor]


def build_model(config: dict[str, Any]) -> tuple[VocalocatorArchitecture, LossFunction]:
    """
    Specifies model and loss funciton.

    Parameters
    ----------
    config : dict
        Dictionary holding training hyperparameters.

    Returns
    -------
    model : VocalocatorArchitecture
        Model instance with hyperparameters specified
        in config.

    loss_function : LossFunction
        Loss function mapping network output to a
        per-instance. That is, loss_function should
        take a torch.Tensor with shape (batch_size, ...)
        and map it to a shape of (batch_size,) holding
        losses.
    """
    # 1. parse desired output type
    output_factory = make_output_factory(config)

    # if the output is probabilistic in nature, use NLL
    if issubclass(output_factory.output_type, ProbabilisticOutput):
        loss = negative_log_likelihood
    # otherwise, just use squared error
    else:
        loss = squared_error

    arch: str = config["ARCHITECTURE"].lower()

    if arch == "vocalocatorensemble":
        # None out the other parameters
        built_submodels = []
        for sub_model_config in config["MODEL_PARAMS"]["CONSTITUENT_MODELS"]:
            if not isinstance(sub_model_config, dict):
                # A path to a config was passed instead of the config contents
                with open(sub_model_config, "r") as f:
                    if not Path(sub_model_config).exists():
                        raise ValueError(
                            f"Path to submodel config {sub_model_config} does not exist!"
                        )
                    sub_model_config = json.load(f)
            submodel, _ = build_model(sub_model_config)
            built_submodels.append(submodel)
        model = VocalocatorEnsemble(config, built_submodels, output_factory)
    elif arch.lower() in ARCHITECTURES:
        model = ARCHITECTURES[arch]
        model = model(config, output_factory)
    else:
        raise ValueError(f"ARCHITECTURE {arch} not recognized.")

    return model, loss


def __apply_affine(locations: np.ndarray, A: np.ndarray, b: np.ndarray):
    """
    Helper function to apply an affine transformation Ax + b
    to a list of vectors.

    Args:
        locations: Array of shape (n_samples, k)
        A: Array of shape (k, k)
        b: Array of shape (k,)
    """
    return (A @ locations[..., None]).squeeze() + b


def unscale_output(
    model_output: np.ndarray, arena_dims: Union[tuple[float, float], np.ndarray]
) -> np.ndarray:
    """
    Transform model output from arbitrary units on the square [-1, 1]^2 to
    the size of the arena, accounting correctly for the shape of the input.

    Note: expects the first dimension of model_output to be the number of
    samples per vocalization.

    This transformation and recentering is given by the affine transform

        z_i = Ay_i + b,

    where z_i is the new location and

        b = 1/2 * (a_1, a_2),

        A = 1/2 * [[a_1 ,  0  ],
                   [0   ,  a_2]].
    """
    A = 0.5 * np.diag(arena_dims)  # rescaling matrix
    b = 0.5 * np.array(arena_dims)  # recentering vector

    unscaled = np.zeros(model_output.shape)

    # if the model outputs a mean y and a covariance matrix S, the transformed
    # mean is given by Ay + b, and the cov matrix is given by A @ S @ A.T
    # (result from the transformation of a random variable)
    if model_output.ndim == 3 and model_output.shape[1:] == (3, 2):
        means = model_output[:, 0]  # shape: (len(model_output), 2)
        cholesky = model_output[:, 1:]  # shape: (len(model_output), 2, 2)
        covs = cholesky @ cholesky.swapaxes(-1, -2)

        unscaled[:, 0] = __apply_affine(means, A, b)
        unscaled[:, 1:] = A @ covs @ A.T

    # similar if model outputs a batch of means + cholesky covariances
    # this is the case for ensemble models
    elif model_output.ndim == 4 and model_output.shape[2:] == (3, 2):
        means = model_output[:, :, 0]  # shape: (batch_size, n_models, 2)
        cholesky = model_output[:, :, 1:]  # shape: (batch_size, n_models, 2, 2)
        covs = cholesky @ cholesky.swapaxes(-1, -2)

        unscaled[:, :, 0] = __apply_affine(means, A, b)
        unscaled[:, :, 1:] = A @ covs @ A.T

    # otherwise, just apply the affine transformation
    elif model_output.ndim == 2 and model_output.shape[1] == 2:
        unscaled = __apply_affine(model_output, A, b)
    else:
        raise ValueError(
            f"Unscaling not currently supported for output of shape {model_output.shape}!"
        )

    return unscaled



import argparse
from os import path
from pathlib import Path

from vocalocator.training.configs import build_config
from vocalocator.training.trainer import Trainer


def get_args():
    parser = argparse.ArgumentParser()

    # Configs can be provided as either a name, which then references an entry in the dictionary
    # located in configs.py, or as a path to a JSON file, when then uses the entries in that file
    # to override the default configuration entries.

    parser.add_argument(
        "--config",
        type=str,
        help="Used to specify model configuration via a JSON file.",
    )

    parser.add_argument(
        "--data",
        type=str,
        help="Path to directory containing train, test, and validation datasets or single h5 file for inference",
    )

    parser.add_argument(
        "--save-path",
        type=Path,
        required=False,
        default=path.abspath("."),
        help="Directory for trained models' weights",
    )

    parser.add_argument(
        "--indices",
        type=Path,
        required=False,
        default=None,
        help="Path to directory containing train/val indices. Should be provided as train_set.npy and val_set.npy",
    )

    args = parser.parse_args()
    validate_args(args)
    return args


def validate_args(args):
    if args.config is None:
        raise ValueError("No config file provided.")

    if not path.exists(args.config):
        raise ValueError(
            f"Requested config JSON file could not be found: {args.config}"
        )

    if args.save_path is None:
        raise ValueError("No save path (trained model storage location) provided.")

    if args.data is None:
        raise ValueError("No data path provided.")

    args.config_data = build_config(args.config)

    Path(args.save_path).mkdir(parents=True, exist_ok=True)


def run(args):
    weights = args.config_data.get("WEIGHTS_PATH", None)

    # This modifies args.config_data['WEIGHTS_PATH']
    trainer = Trainer(
        data_dir=args.data,
        model_dir=args.save_path,
        config_data=args.config_data,
        eval_mode=False,
        index_dir=args.indices,
    )

    if weights is not None:
        trainer.model.load_weights(best_weights_path=weights)

    trainer.train()


if __name__ == "__main__":
    args = get_args()
    run(args)



from vocalocator.outputs.base import (
    MDNOutput,
    ModelOutput,
    PointOutput,
    ProbabilisticOutput,
    UniformOutput,
)
from vocalocator.outputs.factory import ModelOutputFactory
from vocalocator.outputs.gaussian import (
    GaussianOutputDiagonalCov,
    GaussianOutputFixedVariance,
    GaussianOutputFullCov,
    GaussianOutputSphericalCov,
)



"""
File to configure all hyperparameters (model architecture and training).
"""

from sys import stderr
from typing import NewType

try:
    # Attempt to use json5 if available
    import pyjson5 as json
except ImportError:
    print("Warning: json5 not available, falling back to json.", file=stderr)
    import json


JSON = NewType("JSON", dict)
DEFAULT_CONFIG = {
    "OPTIMIZATION": {
        "NUM_EPOCHS": 50,
        "OPTIMIZER": "SGD",
        "MOMENTUM": 0.7,
        "CLIP_GRADIENTS": False,
        "INITIAL_LEARNING_RATE": 0.037,
        "SCHEDULERS": [
            {"SCHEDULER_TYPE": "COSINE_ANNEALING", "MIN_LEARNING_RATE": 0.0}
        ],
    },
    "ARCHITECTURE": "VocalocatorSimpleNetwork",
    "GENERAL": {
        "DEVICE": "GPU",  # 'GPU' or 'CPU'
        "TORCH_SEED": 888,  # rng seeds for reproducibility
        "NUMPY_SEED": 777,
        "LOG_INTERVAL": 3,  # Amount of time between consecutive log messages
    },
    "DATA": {
        "NUM_MICROPHONES": 4,
        "SAMPLE_RATE": 125000,
        "BATCH_SIZE": 32,
        "CROP_LENGTH": 8192,
        "AUGMENT_DATA": True,
        "NORMALIZE_DATA": True,
        "VOCALIZATION_DIR": None,
        "ARENA_DIMS": [572, 356],
        "ARENA_DIMS_UNITS": "MM",
    },
    "AUGMENTATIONS": {
        # Data augmentations: involves performing augmentations to the audio to which the model should be invariant
        "INVERSION": {
            "PROB": 0.5,
        },
        "NOISE": {
            "MIN_SNR": 3,
            "MAX_SNR": 15,
            "PROB": 0.5,
        },
        "MASK": {
            "PROB": 0.5,
            "MIN_LENGTH": 75,  # 0.6 ms at 125 kHz
            "MAX_LENGTH": 125,  # 1 ms at 125 kHz
        },
    },
}


def keys_to_uppercase(dictionary: dict) -> dict:
    """Converts all the string keys in a dictionary to uppercase for consistency"""
    new_dict = dict()
    for key, value in dictionary.items():
        if isinstance(value, dict):  # recurse through subdictionaries if present
            value = keys_to_uppercase(value)
        if isinstance(key, str):
            new_dict[key.upper()] = value
        else:
            new_dict[key] = value
    return new_dict


def update_recursively(dictionary: dict, defaults: dict) -> dict:
    """Updates a dictionary with default values, recursing through subdictionaries"""
    for key, default_value in defaults.items():
        if key not in dictionary:
            dictionary[key] = default_value
        elif isinstance(dictionary[key], dict):
            dictionary[key] = update_recursively(dictionary[key], default_value)
    return dictionary


def build_config(filepath: str) -> JSON:
    with open(filepath, "r") as ctx:
        try:
            config = json.load(ctx)
        except:
            raise ValueError(
                f"Could not parse JSON file at {filepath}. Perhaps a JSON5 file was provided without the necessary libraries installed?"
            )

    config = update_recursively(config, DEFAULT_CONFIG)
    return keys_to_uppercase(config)



from typing import Union

import torch
from vocalocator.outputs.base import EnsembleOutput, MDNOutput, ModelOutput, Unit


class ModelOutputFactory:
    """
    Util class passed to model instances that provides the number of expected
    parameters and a creator method taking a tensor and returning a ModelOutput
    object.
    """

    def __init__(
        self,
        arena_dims: torch.Tensor,
        arena_dim_units: Union[str, Unit],
        output_type: type[ModelOutput],
        **kwargs,
    ):
        self.arena_dims = arena_dims
        self.arena_dim_units = arena_dim_units
        self.output_type = output_type
        self.type_specific_kwargs = kwargs

        # expected_param_count_known = hasattr(self.output_type, 'N_OUTPUTS_EXPECTED')
        if hasattr(self.output_type, "N_OUTPUTS_EXPECTED"):
            self.n_outputs_expected = self.output_type.N_OUTPUTS_EXPECTED
        elif self.output_type == MDNOutput:
            # if supposed to create a mixture density network,
            # we need to parse the kwargs to understand how many
            # outputs are expected
            constituents = self.type_specific_kwargs.get("constituent_response_types")
            if not constituents:
                raise ValueError(
                    "Cannot create MDNOutput without "
                    "keyword argument `constituent_response_types`, which "
                    "should be a list of the base distributions the mixture "
                    "density network should be interpreted as parameterizing."
                )
            # sum the number of parameters each constituent response distribution expects
            n_params_for_constituents = sum(t.N_OUTPUTS_EXPECTED for t in constituents)
            # and expect one value per constituent distribution representing a
            # log of the mixing weights
            n_mixing_params = len(constituents)
            self.n_outputs_expected = n_params_for_constituents + n_mixing_params
        elif self.output_type == EnsembleOutput:
            # ensemble is a bit of a weird exception
            # probably worth cleaning up and readjusting at some point
            # for now, though, it can take a variable number of arguments
            # since it expects an output of type List[ProbabilisticOutput]
            self.n_outputs_expected = None
        else:
            raise ValueError(
                f"ModelOutputFactory doesn't support creating objects of type {self.output_type}. "
                f"Maybe {self.output_type} is only meant to be used as an interface and not instantiated?"
            )

    def create_output(self, x) -> ModelOutput:
        """
        Package a batch of raw model output `x` into an appropriate
        `ModelOutput` instance.
        """
        return self.output_type(
            raw_output=x,
            arena_dims=self.arena_dims,
            arena_dim_units=self.arena_dim_units,
            **self.type_specific_kwargs,
        )



"""
Defines a helpful ProgressLogger class for displaying
training progress in a deep learning setting.
"""
import os
from time import time

import numpy as np


def format_seconds(seconds):
    """
    Converts decimal number of seconds into a nice string.
    """
    days, remainder = divmod(seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    mins, secs = divmod(remainder, 60)

    outstr = f"{secs:02.0f} secs"
    if mins > 0:
        outstr = f"{mins:02g} mins, " + outstr
    if hours > 0:
        outstr = f"{hours:02g} hours, " + outstr
    if days > 0:
        outstr = f"{days:02g} days, " + outstr

    return outstr


class LossAccumulator:
    def log(self, loss, num_correct, batch_size, elapsed):
        self.imagecount += batch_size
        self.total_loss += loss * batch_size
        self.total_correct += num_correct
        self.total_elapsed += elapsed

    def reset(self):
        self.total_correct = 0
        self.total_loss = 0.0
        self.total_elapsed = 0.0
        self.imagecount = 0

    @property
    def secs_per_image(self):
        return self.total_elapsed / self.imagecount

    @property
    def accuracy(self):
        return self.total_correct / self.imagecount

    @property
    def mean_loss(self):
        return self.total_loss / self.imagecount


class ProgressLogger:
    def __init__(
        self, num_epochs, traindata, valdata, log_interval, output_dir, logger
    ):
        self.num_epochs = num_epochs
        self.num_train_images = len(traindata.dataset)
        self.num_val_images = len(valdata.dataset)
        self.log_interval = log_interval
        self.epochcount = 0
        self.epoch_start_time = None
        self.time_ckpt = time()
        self.elapsed_training = 0.0
        self.elapsed_testing = 0.0
        self.last_log_time = -np.inf
        self.time_initialized = time()

        self.train_accumulator = LossAccumulator()
        self.val_accumulator = LossAccumulator()

        self.train_loss_filepath = os.path.join(output_dir, "train_loss.txt")
        self.val_loss_filepath = os.path.join(output_dir, "val_loss.txt")
        self.val_calibration_filepath = os.path.join(output_dir, "val_calibration.txt")

        self.logger = logger

    def require_log(self):
        """Returns True if we need to log progress."""
        if (time() - self.last_log_time) > self.log_interval:
            self.last_log_time = time()
            return True
        else:
            return False

    def timer_split(self):
        """Returns time in seconds between each function call."""
        split = time() - self.time_ckpt
        self.time_ckpt = time()
        return split

    def start_training(self):
        # Increment epochs.
        self.epochcount += 1
        self.logger.info(">> STARTING EPOCH {}".format(self.epochcount))

        # Print estimated overall time remaining.
        if self.epochcount > 1:
            epochs_left = 1 + self.num_epochs - self.epochcount
            time_remaining = epochs_left * (
                self.train_accumulator.secs_per_image * self.num_train_images
                + self.val_accumulator.secs_per_image * self.num_val_images
            )
            self.logger.info(
                ">> TIME ELAPSED SO FAR:\t" + self.print_time_since_initialization()
            )
            self.logger.info(
                ">> EST. TIME REMAINING:\t" + format_seconds(time_remaining)
            )

        # Reset training statistics.
        self.train_accumulator.reset()

        # Used to print total time elapsed per epoch.
        self.epoch_start_time = time()

        # Reset timer before commencing training.
        self.timer_split()

    def log_train_batch(self, loss, num_correct, batch_size):
        # Log training statistics.
        self.train_accumulator.log(loss, num_correct, batch_size, self.timer_split())

        # Output training progress.
        if self.require_log():
            self.logger.info(
                "TRAINING. \t Epoch "
                + "{} / {} ".format(self.epochcount, self.num_epochs)
                + "[{}/{}]".format(
                    self.train_accumulator.imagecount, self.num_train_images
                )
                + f"   minibatch loss: {loss:.5f}"
            )

    def start_testing(self):
        # Log progress.
        self.logger.info(">> DONE TRAINING, STARTING TESTING.")

        # Reset testing statistics.
        self.val_accumulator.reset()

        # Reset time before commencing testing.
        self.timer_split()

    def log_val_batch(self, loss, num_correct, batch_size):
        # Log testing statistics.
        self.val_accumulator.log(loss, num_correct, batch_size, self.timer_split())
        # Output training progress.
        if self.require_log():
            self.logger.info(
                "TESTING VALIDATION SET. Epoch {} ".format(self.epochcount)
                + "[{}/{}]".format(self.val_accumulator.imagecount, self.num_val_images)
            )

    def finish_epoch(self, calibration_curve=None):
        """Log statistics on the test set."""
        self.logger.info(
            ">> FINISHED EPOCH IN: " + format_seconds(time() - self.epoch_start_time)
        )

        train_loss = self.train_accumulator.mean_loss
        val_loss = self.val_accumulator.mean_loss

        with open(self.train_loss_filepath, "a") as f:
            f.write(f"{train_loss}\n")

        with open(self.val_loss_filepath, "a") as f:
            f.write(f"{val_loss}\n")

        if calibration_curve is not None:
            with open(self.val_calibration_filepath, "a") as f:
                f.write(f"{calibration_curve}\n")

        return val_loss

    def print_time_since_initialization(self):
        """Display time since ProgressLogger was initialized."""
        return format_seconds(time() - self.time_initialized)
